/* io.c - I/O operations */
 
/* Written 1995-1998 by Werner Almesberger, EPFL-LRC/ICA */
 

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include "../linux-atm.h"
#include "../qgen/uni.h"

#include "atm.h"
#include "atmd.h"
#include "localnsap.h"

#include "atmarpd.h"
#include "table.h"
#include "arp.h"
#include "itf.h"
#include "io.h"
#include "unix.h"


#define COMPONENT "IO"


struct timeval now;


static int kernel, incoming, inet, unix_sock;


/* ----- kernel interface -------------------------------------------------- */


static void open_kernel(void)
{
    struct svc_addr addr;
    struct atm_qos qos;
    struct atm_sap sap;

#ifdef todo
    if ((kernel = socket(PF_ATMSVC, SOCK_DGRAM, 0)) < 0)
	diag(COMPONENT, DIAG_FATAL, "socket: %s", strerror(errno));
    if ((incoming = socket(PF_ATMSVC, SOCK_DGRAM, 0)) < 0)
	diag(COMPONENT, DIAG_FATAL, "socket: %s", strerror(errno));
#endif
    memset(&qos, 0, sizeof(qos));
    qos.aal = ATM_AAL5;
    qos.rxtp.traffic_class = qos.txtp.traffic_class = ATM_ANYCLASS;
#ifdef todo
    if (setsockopt(incoming, SOL_ATM, SO_ATMQOS, &qos, sizeof(qos)) < 0)
	diag(COMPONENT, DIAG_FATAL,
	     "setsockopt SO_ATMQOS: %s", strerror(errno));
#endif
    memset(&sap, 0, sizeof(sap));
    sap.blli[0].l2_proto = ATM_L2_ISO8802;
    sap.blli[0].l3_proto = ATM_L3_NONE;
#ifdef todo
    if (setsockopt(incoming, SOL_ATM, SO_ATMSAP, &sap, sizeof(sap)) < 0)
	diag(COMPONENT, DIAG_FATAL,
	     "setsockopt SO_ATMSAP: %s", strerror(errno));
    memset(&addr, 0, sizeof(addr));
    if (bind(incoming, (struct sockaddr *) &addr, sizeof(addr)) >= 0) {
	if (listen(incoming, 5) < 0)
	    diag(COMPONENT, DIAG_FATAL, "listen: %s", strerror(errno));
    } else {
	if (errno != EUNATCH)
	    diag(COMPONENT, DIAG_FATAL, "bind: %s", strerror(errno));
	diag(COMPONENT, DIAG_WARN, "SVCs are not available");
	(void) close(incoming);
	incoming = -1;
    }
#endif
    if ((inet = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
	diag(COMPONENT, DIAG_FATAL, "socket: %s", strerror(errno));
}


static void recv_kernel(void)
{
#ifdef todo
    struct atmarp_ctrl ctrl;
    int size;

    size = read(kernel, &ctrl, sizeof(ctrl));
    if (size < 0) {
	diag(COMPONENT, DIAG_ERROR, "read kernel: %s", strerror(errno));
	return;
    }
    switch (ctrl.type) {
	case act_need:
	    need_ip(ctrl.itf_num, ctrl.ip);
	    break;
	case act_up:
	    itf_up(ctrl.itf_num);
	    break;
	case act_down:
	    itf_down(ctrl.itf_num);
	    break;
	case act_change:
	    itf_change(ctrl.itf_num);
	default:
	    diag(COMPONENT, DIAG_ERROR,
		 "invalid control msg type 0x%x", ctrl.type);
    }
#endif
}


static void close_kernel(void)
{
    if (incoming >= 0)
	(void) close(incoming);
    (void) close(kernel); /* may get major complaints from the kernel ... */
    (void) close(inet);
}


/* ----- atmarp (maintenance) interface ------------------------------------ */


static void open_unix(void)
{
    unix_sock = un_create(ATMARP_SOCKET_PATH, 0600);
    if (unix_sock < 0)
	diag(COMPONENT, DIAG_FATAL, "un_create: %s", strerror(errno));
}


static int got_unix(void *buf, int len, void *user)
{
    struct atmarp_req *req = buf;
    int *reply = buf;

    if (len != sizeof(*req)) {
	diag(COMPONENT, DIAG_ERROR,
	     "bad unix read: %d != %d", len, sizeof(*req));
	return 0;
    }
    switch (req->type) {
	case art_create:
#ifdef todo
	    *reply = ioctl(kernel, SIOCMKCLIP, req->itf);
#endif
	    if (*reply >= 0)
		itf_create(*reply);
	    break;
	case art_qos:
	case art_set:
	case art_delete:
	    *reply = arp_ioctl(req);
	    break;
	case art_table:
	    *reply = table_update();
	    break;
	default:
	    diag(COMPONENT, DIAG_ERROR, "invalid request msg type 0x%x",
	      req->type);
	    *reply = -EINVAL;
    }
    return sizeof(int);
}


static void recv_unix(void)
{
    struct atmarp_req req;

    if (un_reply(unix_sock, &req, sizeof(req), got_unix, NULL) < 0)
	diag(COMPONENT, DIAG_ERROR, "un_reply: %s", strerror(errno));
}


static void close_unix(void)
{
    (void) close(unix_sock);
    (void) unlink(ATMARP_SOCKET_PATH);
}


/* ----- common part ------------------------------------------------------- */


#define MAX_BUFFER 1024


static fd_set rset, cset;


int do_close(int fd)
{
    int result;

    result = close(fd);
    /* we might open a new fd with the same number, so ... */
    FD_CLR(fd, &rset);
    FD_CLR(fd, &cset);
    return result;
}


static void recv_vcc(VCC *vcc)
{
    u_char buffer[MAX_BUFFER];
    int size;

    size = read(vcc->fd, buffer, MAX_BUFFER);
    if (size == 0) {
	disconnect_vcc(vcc);
	return;
    }
    if (size < 0) {
	diag(COMPONENT, DIAG_ERROR, "read vcc: %s", strerror(errno));
	disconnect_vcc(vcc);
	return;
    }
    if (debug) {
	int i;

	for (i = 0; i < size; i++)
	    printf("%02X ", buffer[i]);
	printf("\n");
    }
    incoming_arp(vcc, (struct atmarphdr *) buffer, size);
}


static void accept_new(void)
{
    char buffer[MAX_ATM_ADDR_LEN + 1];
    struct svc_addr addr;
    ENTRY *entry;
    VCC *vcc;
    int fd, len, error;

    len = sizeof(addr);
#ifdef todo
    if ((fd = accept(incoming, (struct sockaddr *) &addr, &len)) < 0) {
	error = errno;
	diag(COMPONENT, DIAG_ERROR, "accept: %s", strerror(errno));
	if (error == EUNATCH) {
	    diag(COMPONENT, DIAG_WARN, "disabling SVCs");
	    (void) close(incoming);
	    incoming = -1;
	}
	return;
    }
    if (ioctl(fd, ATMARP_MKIP, CLIP_DEFAULT_IDLETIMER) < 0) {
        diag(COMPONENT, DIAG_ERROR, "ioctl ATMARP_MKIP: %s", strerror(errno));
        (void) do_close(fd);
        return;
    }
#endif
    vcc = alloc_t(VCC);
    vcc->active = 0;
    vcc->connecting = 0;
    vcc->fd = fd;
    /* the following code probably belongs to arp.c ... */
    if (atm2text_svc(buffer, MAX_ATM_ADDR_LEN + 1, &addr, pretty) < 0)
	strcpy(buffer, "<atm2text error>");
    diag(COMPONENT, DIAG_DEBUG, "Incoming call from %s", buffer);
    if (merge) {
	ITF *itf;

	for (itf = itfs; itf; itf = itf->next) {
	    entry = lookup_addr(itf, &addr);
	    if (entry) {
		vcc->entry = entry;
		Q_INSERT_HEAD(entry->vccs, vcc);
		if (entry->state == as_valid)
		    if (set_ip(vcc->fd, entry->ip) < 0)
			diag(COMPONENT, DIAG_ERROR, "set_ip: %s",
			     strerror(errno));
		return;
	    }
	}
    }
    entry = alloc_entry(1);
    entry->state = as_invalid;
    entry->addr = alloc_t(struct svc_addr);
    *entry->addr = addr;
    entry->flags = ATF_PUBL;
    Q_INSERT_HEAD(unknown_incoming, entry);
    vcc->entry = entry;
    Q_INSERT_HEAD(entry->vccs, vcc);
    incoming_call(vcc);
}


int connect_vcc(struct svc_addr *remote,
		const struct atm_qos *qos,
		int timeout)
{
    int fd, flags;
    struct atm_sap sap;

    if (incoming < 0)
	return -ENOTCONN;
#ifdef todo
    if ((fd = socket(remote->sa_family, SOCK_DGRAM, 0)) < 0) {
	diag(COMPONENT, DIAG_ERROR, "socket: %s", strerror(errno));
	return -errno;
    }
    if (setsockopt(fd, SOL_ATM, SO_ATMQOS, qos, sizeof(*qos)) < 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "setsockopt SO_ATMQOS: %s", strerror(errno));
	return -errno;
    }
#endif
    if ((flags = fcntl(fd, F_GETFL)) < 0) {
	diag(COMPONENT, DIAG_ERROR, "fcntl F_GETFL: %s", strerror(errno));
	return -errno;
    }
    flags |= O_NONBLOCK;
    if (fcntl(fd, F_SETFL, flags) < 0) {
	diag(COMPONENT, DIAG_ERROR, "fcntl F_GETFL: %s", strerror(errno));
	return -errno;
    }
    memset(&sap, 0, sizeof(sap));
    sap.blli[0].l2_proto = ATM_L2_ISO8802;
    sap.blli[0].l3_proto = ATM_L3_NONE;
#ifdef todo
    if (setsockopt(fd, SOL_ATM, SO_ATMSAP, &sap, sizeof(sap)) < 0) {
	diag(COMPONENT, DIAG_ERROR, "setsockopt SO_ATMSAP: %s",
	     strerror(errno));
	return -errno;
    }
    if (ioctl(fd, ATMARP_MKIP, timeout) < 0) {
        diag(COMPONENT, DIAG_ERROR, "ioctl ATMARP_MKIP: %s", strerror(errno));
        (void) do_close(fd);
        return -errno;
    }
#endif
    return fd;
}


int set_ip(int fd, int ip)
{
#ifdef todo
    if (ioctl(fd, ATMARP_SETENTRY, ip) >= 0)
	return 0;
#endif
    diag(COMPONENT, DIAG_ERROR, "ioctl ATMARP_SETENTRY: %s", strerror(errno));
    (void) do_close(fd);
    return -errno;
}


int set_encap(int fd, int mode)
{
#ifdef todo
    if (ioctl(fd, ATMARP_ENCAP, mode) >= 0)
	return 0;
#endif
    diag(COMPONENT, DIAG_ERROR, "ioctl ATMARP_ENCAP: %s", strerror(errno));
    (void) do_close(fd);
    return -errno;
}


static void complete_connect(VCC *vcc)
{
    struct svc_addr dummy;

    if (!vcc->connecting)
	diag(COMPONENT, DIAG_FATAL, "connecting non-connecting VCC 0x%p", vcc);
    memset(&dummy, 0, sizeof(dummy));
#ifdef todo
    if (!connect(vcc->fd, (struct sockaddr *) &dummy, sizeof(dummy))) {
	if (ioctl(vcc->fd, ATMARP_MKIP, CLIP_DEFAULT_IDLETIMER) < 0) {
	    diag(COMPONENT, DIAG_ERROR,
		 "ioctl ATMARP_MKIP: %s", strerror(errno));
	    (void) do_close(vcc->fd);
	    vcc_failed(vcc);
	}
	vcc_connected(vcc);
    } else {
	diag(COMPONENT, DIAG_INFO, "[2]connect: %s", strerror(errno));
	(void) do_close(vcc->fd);
	vcc_failed(vcc);
    }
#endif
}


void poll_loop(void)
{
    ITF *itf, *next_itf;
    ENTRY *entry, *next_entry;
    VCC *vcc, *next_vcc;
    int fds, ret;

    gettimeofday(&now, NULL);
    while (1) {
	FD_ZERO(&rset);
	FD_ZERO(&cset);
	FD_SET(kernel, &rset);
	FD_SET(unix_sock, &rset);
	if (incoming >= 0)
	    FD_SET(incoming, &rset);
	fds = incoming + 1;
	if (kernel >= fds)
	    fds = kernel + 1;
	if (unix_sock >= fds)
	    fds = unix_sock + 1;
	for (itf = itfs; itf; itf = itf->next)
	    for (entry = itf->table; entry; entry = entry->next)
		for (vcc = entry->vccs; vcc; vcc = vcc->next) {
		    if (vcc->connecting)
			FD_SET(vcc->fd, &cset);
		    else
			FD_SET(vcc->fd, &rset);
		    if (vcc->fd >= fds)
			fds = vcc->fd+1;
		}
	for (entry = unknown_incoming; entry; entry = entry->next) {
	    if (!entry->vccs || entry->vccs->next) {
		diag(COMPONENT, DIAG_ERROR,
		     "internal error: bad unknown entry");
		continue;
	    }
	    FD_SET(entry->vccs->fd, &rset);
	    if (entry->vccs->fd >= fds)
		fds = entry->vccs->fd+1;
	}
	ret = select(fds, &rset, &cset, NULL, next_timer());
	if (ret < 0) {
	    if (errno != EINTR)
		perror("select");
	} else {
	    diag(COMPONENT, DIAG_DEBUG, "----------");
	    gettimeofday(&now, NULL);
	    if (FD_ISSET(kernel, &rset))
		recv_kernel();
	    if (FD_ISSET(unix_sock, &rset))
		recv_unix();
	    if (incoming >= 0 && FD_ISSET(incoming, &rset))
		accept_new();
	    for (itf = itfs; itf; itf = next_itf) {
		next_itf = itf->next;
		for (entry = itf->table; entry; entry = next_entry) {
		    next_entry = entry->next;
		    for (vcc = entry->vccs; vcc; vcc = next_vcc) {
			next_vcc = vcc->next;
			if (FD_ISSET(vcc->fd, &rset))
			    recv_vcc(vcc);
			else if (FD_ISSET(vcc->fd, &cset))
			    complete_connect(vcc);
		    }
		}
	    }
	    for (entry = unknown_incoming; entry; entry = next_entry) {
		next_entry = entry->next;
		if (FD_ISSET(entry->vccs->fd, &rset))
		    recv_vcc(entry->vccs);
	    }
	    expire_timers();
	      /* expire timers after handling messages to make sure we don't
		 time out unnecessarily because of scheduling delays */
	}
	table_changed();
    }
}


void send_packet(int fd, void *data, int length)
{
    int wrote;

    if (debug) {
	int i;

	for (i = 0; i < length; i++)
	    printf("%02X ", ((u_char *) data)[i]);
	printf("\n");
    }
    if ((wrote = write(fd, data, length)) == length)
	return;
    if (wrote < 0)
	diag(COMPONENT, DIAG_ERROR, "write: %s", strerror(errno));
    else
	diag(COMPONENT, DIAG_ERROR, "short write: %d < %d", wrote, length);
}


int ip_itf_info(int number, uint32_t *ip, uint32_t *netmask, int *mtu)
{
    struct ifreq req;
    u_char *p1, *p2;

    memset(&req, 0, sizeof(req));
    sprintf(req.ifr_name, ENNAME, number);
    if (ioctl(inet, SIOCGIFADDR, &req) < 0) {
	diag(COMPONENT, DIAG_ERROR, "ioctl SIOCGIFADDR: %s", strerror(errno));
	return -1;
    }
    *ip = ((struct sockaddr_in *) &req.ifr_addr)->sin_addr.s_addr;
    if (ioctl(inet, SIOCGIFNETMASK, &req) < 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "ioctl SIOCGIFNETMASK: %s", strerror(errno));
	return -1;
    }
    *netmask = ((struct sockaddr_in *) &req.ifr_addr)->sin_addr.s_addr;
    if (ioctl(inet, SIOCGIFMTU, &req) < 0) {
	diag(COMPONENT, DIAG_ERROR, "ioctl SIOCGIFMTU: %s", strerror(errno));
	return -1;
    }
    *mtu = req.ifr_mtu;
    p1 = (u_char *) ip;
    p2 = (u_char *) netmask;
    diag(COMPONENT, DIAG_DEBUG, "ip %d.%d.%d.%d mask %d.%d.%d.%d mtu %d",
	 p1[0], p1[1], p1[2], p1[3], p2[0], p2[1], p2[2], p2[3], *mtu);
    return 0;
}


int get_local(int fd, struct svc_addr *addr)
{
    get_local_addr(addr, 1, 0);
    return 0;
}


void open_all(void)
{
    open_kernel();
    open_unix();
}


void close_all(void)
{
    close_kernel();
    close_unix();
}
