/* table.c - ATMARP table */
 
/* Written 1995-1998 by Werner Almesberger, EPFL-LRC/ICA */
 

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "../linux-atm.h"

#include "atm.h"
#include "atmd.h"

#include "atmarpd.h"
#include "table.h"


#define COMPONENT "TABLE"


const char *entry_state_name[] = { "NONE", "RESOLV", "INVALID", "VALID" };


ENTRY *alloc_entry(int svc)
{
    ENTRY *entry;

    entry = alloc_t(ENTRY);
    memset(entry, 0, sizeof(ENTRY));
    entry->state = as_none;
    entry->svc = svc;
    return entry;
}


ENTRY *lookup_ip(const ITF *itf, uint32_t ip)
{
    ENTRY *walk;

    for (walk = itf->table; walk; walk = walk->next)
	if (walk->ip == ip)
	    break;
    return walk;
}


ENTRY *lookup_addr(const ITF *itf, const struct svc_addr *addr)
{
    ENTRY *walk;

    for (walk = itf->table; walk; walk = walk->next)
	if (walk->addr && atm_equal(walk->addr, addr, 0, 0))
	    break;
    return walk;
}


ENTRY *lookup_incoming(const struct svc_addr *addr)
{
    ENTRY *walk;

    for (walk = unknown_incoming; walk; walk = walk->next)
	if (walk->addr && atm_equal(walk->addr, addr, 0, 0))
	    break;
    return walk;
}


static int table_uptodate = 0; /* ATMARP table file is up to date */
static FILE *out_file = NULL;
static int out_error = 0;


static void output(const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    if (!out_file)
	vdiag(COMPONENT, DIAG_DEBUG, fmt, ap);
    else if (vfprintf(out_file, fmt, ap) < 0 || putc('\n', out_file) < 0)
	out_error = errno;
    va_end(ap);
}


static void dump_vcc(VCC *vcc)
{
    struct sockaddr_natm addr;
    char buffer[MAX_ATM_ADDR_LEN + 1];
    int size;

    size = sizeof(addr);
    if (getpeername(vcc->fd, (struct sockaddr *) &addr, &size) < 0) {
	diag(COMPONENT, DIAG_ERROR, "getpeername: %s", strerror(errno));
	strcpy(buffer, "<getsocknam error>");
    } else {
	if (atm2text_pvc(buffer, MAX_ATM_ADDR_LEN + 1, &addr, pretty) < 0)
	    strcpy(buffer, "<atm2text error>");
    }
    output("  %s%s", buffer,
	   vcc->connecting ? ",  connecting" :
	      !vcc->entry->svc ? "" :
		  vcc->active ? " (active)" : " (passive)");
}


static void dump_entries(ENTRY *list)
{
    static const char *flag_name[] = {
	"???",	"com", "PERM", "PUBL",			/* 0x0001-0x0008 */
	"trailers", "netmask", "dontpub", "magic",	/* 0x0010-0x0080 */
	"???", "???", "???", "???",			/* 0x0100-0x0800 */
	"NULL", "ARPSRV", "???" };			/* 0x1000-0x8000 */
    /* lower case flags are not used by ATMARP */
    ENTRY *entry;
    VCC *vcc;
    char addr_buf[MAX_ATM_ADDR_LEN + 1];
    char qos_buf[MAX_ATM_QOS_LEN + 1];
    char tmp[100]; /* large enough for all flags */
    u_char *ipp;
    int i;

    for (entry = list; entry ; entry = entry->next) {
	if (!entry->addr)
		strcpy(addr_buf, "<none>");
	else if (atm2text_svc(addr_buf, MAX_ATM_ADDR_LEN + 1,
			      entry->addr, pretty) < 0)
	    strcpy(addr_buf, "<error>");
	ipp = (u_char *) &entry->ip;
	*tmp = 0;
	for (i = 0; i < 16; i++)
	    if (entry->flags & (1 << i)) {
		if (*tmp)
		    strcat(tmp, ",");
		strcat(tmp, flag_name[i]);
	    }
	output("IP %d.%d.%d.%d, state %s, addr %s, flags 0x%x<%s>",
	       ipp[0], ipp[1], ipp[2], ipp[3],
	       entry_state_name[entry->state],
	       addr_buf, entry->flags, tmp);
	if (entry->itf && !qos_equal(&entry->itf->qos, &entry->qos)) {
	    if (qos2text(qos_buf, sizeof(qos_buf), &entry->qos, 0) < 0)
		strcpy(qos_buf, "<error>");
	    else
		output("  QOS: %s", qos_buf);
	}
	for (vcc = entry->vccs; vcc; vcc = vcc->next)
	    dump_vcc(vcc);
    }
}


static void dump_itf(ITF *itf)
{
    u_char *ipp, *nmp;
    char buf[MAX_ATM_QOS_LEN + 1];

    if (qos2text(buf, sizeof(buf), &itf->qos, 0) < 0)
	strcpy(buf, "<error>");
    ipp = (u_char *) &itf->local_ip;
    nmp = (u_char *) &itf->netmask;
    output("----- Itf %d (%d.%d.%d.%d, netmask %d.%d.%d.%d) -----",
	   itf->number,
	   ipp[0], ipp[1], ipp[2], ipp[3],
	   nmp[0], nmp[1], nmp[2], nmp[3]);
    output("Default QOS: %s", buf);
    dump_entries(itf->table);
}


static void dump_all(void)
{
    ITF *itf;

    for (itf = itfs; itf; itf = itf->next)
	dump_itf(itf);
    output("----- Unknown incoming connections -----");
    dump_entries(unknown_incoming);
    output("----- End of dump -----");
    if (out_file) {
    }
}


void table_changed(void)
{
    table_uptodate = 0;
    table_update(); /* @@@ sigh, fix this later */
    if (debug) {
	out_file = 0;
	dump_all();
    }
}


int table_update(void)
{
    if (table_uptodate)
	return 0;
    out_file = fopen(ATMARP_TMP_DUMP_FILE, "w");
    out_error = 0;
    dump_all();
    if (fclose(out_file) < 0)
	out_error = errno;
    if (!out_error)
	if (rename(ATMARP_TMP_DUMP_FILE, ATMARP_DUMP_FILE) < 0)
	    out_error = errno;
	else
	    table_uptodate = 1;
    unlink(ATMARP_TMP_DUMP_FILE);
    return out_error;
}
