/*
 * io.c - Ilmi input/output routines
 *
 * Written by Scott W. Shumate
 * 
 * Copyright (c) 1995-97 All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this
 * software and its documentation is hereby granted,
 * provided that both the copyright notice and this
 * permission notice appear in all copies of the software,
 * derivative works or modified versions, and any portions
 * thereof, that both notices appear in supporting
 * documentation, and that the use of this software is
 * acknowledged in any publications resulting from using
 * the software.
 * 
 * I ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION AND DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS
 * SOFTWARE.
 */

/* Last modified by M. Souissi 09/17/97  */

#include <sys/types.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_atm.h>
#include <net/if_dl.h>
#include <net/route.h>

#include <string.h>
#include <stdlib.h>


#include <netinet/in.h>
#include <netnatm/natm.h>
#include "io.h"
#include "atmd.h"
#include "../linux-atm.h"
#include "atm.h"
#include "localnsap.h"

#define SNMP_VCI 16
#define COMPONENT "IO"
#define MAX_EXTRA_ADDRS 4 /* maximum number of additional addresses that can
			     be manually configured (after ilmid has
			     registered the "official" address) - HACK */

static short atm_itf = -1; /* bad value */

AsnOid *get_esi(int fd, int itf)
{
  static AsnOid *name;
  caddr_t esip;
  int m, n, size;
  
  if (itf != 0)
    diag(COMPONENT, DIAG_FATAL, "ATM_GETESI: unsupported itf=%d", itf);

  if (!(esip = get_addr_esi(itf)))
    diag(COMPONENT, DIAG_FATAL, "ATM_GETESI: %s", strerror(errno));
    
  /* Convert hex string to Object ID BER */
  for(m = 0, size = 0; m < ESI_LEN; esip[m++] & 0x80 ? size += 2 : size++);
  size++;
  name = alloc_t(AsnOid);
  name->octs = alloc(size);
  name->octetLen = size;
  
  for(m = 0, n = 0; m < ESI_LEN; m++, n++)
    {
      if(esip[m] & 0x80)
	name->octs[n++] = '\201';
      name->octs[n] = esip[m] & 0x7F;
    }
  
  /* Add the SEL */
  name->octs[n] = '\0';
 
  return name;
}

void update_nsap(int itf, AsnOid *netprefix, AsnOid *esi)
{
    FILE *file;
    char fname[64];
    struct svc_addr addr, ouraddr[MAX_EXTRA_ADDRS + 1];
    char buffer[MAX_ATM_ADDR_LEN + 1];
    int m, n;

    memset(&addr, 0, sizeof(addr));
  
    /* Convert net prefix BER to hex */
    for (m = 0, n = 0; m < netprefix->octetLen; m++, n++)
	if (netprefix->octs[m] & 0x80)
	    addr.prv[n] = netprefix->octs[++m] | 0x80;
	else
	    addr.prv[n] = netprefix->octs[m];

    /* Convert esi BER to hex */
    for (m = 0; m < esi->octetLen; m++, n++)
	if (esi->octs[m] & 0x80)
	    addr.prv[n] = esi->octs[++m] | 0x80;
	else
	    addr.prv[n] = esi->octs[m];

    /* Get our address(es) on that interface */

    get_local_addr(&ouraddr[0], MAX_EXTRA_ADDRS + 1, itf);

    if (atmsvc_addr_in_use(ouraddr[0]) &&
	atm_equal(&addr, &ouraddr[0], ATM_ESA_LEN, 0)) {
	diag(COMPONENT, DIAG_INFO, "Primary ATM Address did not change");
	return;
    }

    /* Prepend our address in front */

    (void) snprintf(fname, sizeof(fname), NSAPFILE, itf);
    file = fopen(fname, "w");
    if (!file)
	diag(COMPONENT, DIAG_FATAL, "Can't update ATM Address in %s", fname);

    if (atm2text_svc(buffer, MAX_ATM_ADDR_LEN + 1, &addr,
		     A2T_PRETTY | A2T_LOCAL) < 0)
	diag(COMPONENT, DIAG_FATAL, "Bad Primary ATM Address?");
    fprintf(file, "%s\n", buffer);
    diag(COMPONENT, DIAG_INFO, "Primary ATM Address %s added local", buffer);

    /* Put extra addresses */

    for (n = 0; n < MAX_EXTRA_ADDRS; n++) {
	
	/* Finished ? */
	if (!atmsvc_addr_in_use(ouraddr[n]))
	    break;

	if (atm2text_svc(buffer, MAX_ATM_ADDR_LEN + 1, &ouraddr[n],
			 A2T_PRETTY | A2T_LOCAL) < 0) {
	    diag(COMPONENT, DIAG_ERROR, "Bad Extra ATM Address[%d]?", n);
	    continue;
	}
	fprintf(file, "%s\n", buffer);
	diag(COMPONENT, DIAG_INFO, "Extra ATM Address %s added local", buffer);
    }
    (void) fclose(file);
}

int wait_for_message(int fd, struct timeval *timeout)
{
  int numfds;
  fd_set fdvar;

  FD_ZERO(&fdvar);
  FD_SET(fd, &fdvar);

  if((numfds = select(fd + 1, &fdvar, 0, 0, timeout)) < 0)
    diag(COMPONENT, DIAG_FATAL, "select: %s", strerror(errno));

  return numfds;
}


int read_message(int fd, Message *message)
{
  SBuf buffer;
  char data[MAX_ILMI_MSG];
  AsnLen length;
  jmp_buf env;

  if ((int) (length = read(fd, data, MAX_ILMI_MSG)) < 0)
    diag(COMPONENT, DIAG_FATAL, "read: %s", strerror(errno));

  SBufInstallData(&buffer, data, length);
  if (setjmp(env) == 0)
    {
      BDecMessage(&buffer, message, &length, env);
    }
  else
    {
      diag(COMPONENT, DIAG_ERROR, "message decoding error");
      return -1;
    }

  diag(COMPONENT, DIAG_DEBUG, "SNMP message received:");
  if(get_verbosity(NULL) == DIAG_DEBUG)
    PrintMessage(get_logfile(), message, 0);
  
  if(message->version != VERSION_1)
    {
      diag(COMPONENT, DIAG_ERROR, "received message with wrong version number");
      return -1;
    }

  if(message->community.octetLen != 4 || 
     memcmp(message->community.octs, "ILMI", 4))
    {
      diag(COMPONENT, DIAG_ERROR, "received message with wrong community");
      return -1;
    }

  return 0;
}

int send_message(int fd, Message *message)
{
  SBuf buffer;
  AsnLen length;
  char data[MAX_ILMI_MSG];

  SBufInit(&buffer, data, MAX_ILMI_MSG);
  SBufResetInWriteRvsMode(&buffer);
  
  if(!(length = BEncMessage(&buffer, message)))
  {
    diag(COMPONENT, DIAG_ERROR, "message encoding error");
    return -1;
  }

  if(write(fd, SBufDataPtr(&buffer), length) != length)
    diag(COMPONENT, DIAG_FATAL, "write: %s", strerror(errno));

  diag(COMPONENT, DIAG_DEBUG, "SNMP message sent:");
  if(get_verbosity(NULL) == DIAG_DEBUG)
    PrintMessage(get_logfile(), message, 0);
  
  return 0;
}


int open_ilmi(int itf, const char *qos_spec)
{
    struct sockaddr_natm addr;
    struct atm_qos qos;
    int fd;

    if ((fd = socket(PF_NATM, SOCK_STREAM, PROTO_NATMAAL5)) < 0)
	diag(COMPONENT, DIAG_FATAL, "socket: %s", strerror(errno));

    memset(&qos, 0, sizeof(qos));
    qos.rxtp.max_sdu = MAX_ILMI_MSG;
    qos.txtp.max_sdu = MAX_ILMI_MSG;
    qos.aal = ATM_AAL5;
    if (!qos_spec)
	qos.rxtp.traffic_class = qos.txtp.traffic_class = ATM_UBR;
    else if (text2qos(qos_spec, &qos, T2Q_DEFAULTS) < 0)
	diag(COMPONENT, DIAG_FATAL, "invalid qos: %s", qos_spec);

    memset(&addr, 0, sizeof(addr));
    addr.snatm_len = sizeof(addr);
    addr.snatm_family = AF_NATM;
    snprintf(addr.snatm_if, IFNAMSIZ, ENNAME, itf);
    addr.snatm_vpi = 0;
    addr.snatm_vci = SNMP_VCI;

    if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0)
	diag(COMPONENT, DIAG_FATAL, "connect: %s", strerror(errno));

    if (qos.txtp.max_pcr) {
	struct vctxreq vreq;

	memset(&vreq, 0, sizeof(vreq));
	snprintf(vreq.vc_ifname, IFNAMSIZ, ENNAME, itf);
	vreq.vc_pcr = qos.txtp.max_pcr;
	if (ioctl(fd, SIOCTXSPEED, &vreq) < 0)
	    diag(COMPONENT, DIAG_FATAL,
		 "ioctl SIOCTXSPEED: %s", strerror(errno));
    }
    return fd;
}
