/* atm2text.c - Converts binary encoding of ATM address to textual
		representation */

/* Written 1995-1998 by Werner Almesberger, EPFL-LRC/ICA */


#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <stdio.h>
#include <string.h>
#include "../linux-atm.h"
#include "atm.h"
#include "atmres.h"

static int put_item(char **buffer, int *length, int value)
{
    char *walk, *scan;
    char tmp;
  
    if (!*length)
	return FATAL;
    if (!value) {
	*(*buffer)++ = '0';
	(*length)--;
	return 0;
    }
    for (walk = *buffer; value; value /= 10) {
	if (!(*length)--)
	    return FATAL;
	*walk++ = '0' + (value % 10);
    }
    for (scan = walk - 1; scan > *buffer; (*buffer)++) {
	tmp = *scan;
	*scan-- = **buffer;
	**buffer = tmp;
    }
    *buffer = walk;
    return 0;
}

static int do_pvc(char *buffer, int length, const struct sockaddr_natm *addr,
		  int flags)
{
    int orig_len;
  
    orig_len = length;
    if (put_item(&buffer, &length, addr->snatm_vpi))
	return FATAL;
    if (!length--)
	return FATAL;
    *buffer++ = '.';
    if (put_item(&buffer, &length, addr->snatm_vci))
	return FATAL;
    if (!length)
	return FATAL;
    *buffer = 0;
    return orig_len - length;
}

static int do_svc(char *buffer, int length, const struct svc_addr *addr, 
		  int flags)
{
    static int pure[] = { 20 };
    static int bin[] = { 1,2,10,6,1 };
    static int e164[] = { 4,6,1 };

    int orig_len, len, i, left, value;
    int *fmt;

    orig_len = length;
    if (!*addr->pub && !*addr->prv)
	return FATAL;
    if (*addr->pub) {
	len = strlen(addr->pub);
	if (!*addr->prv && length >= len + 2) {
	    *buffer++ = '+';
	    length--;
	}
	if (length < len+1)
	    return FATAL;
	strcpy(buffer, addr->pub);
	buffer += len;
	length -= len;
	if (*addr->prv) {
	    if (!length--)
		return FATAL;
	    *buffer++ = '+';
	}
    }
    if (*addr->prv) {
	fmt = pure;
	i = 0;
	if (flags & A2T_PRETTY)
	    switch (*addr->prv) {
		case ATM_AFI_DCC:
		case ATM_AFI_ICD:
		    fmt = bin;
		    break;
		case ATM_AFI_E164:
		    for (i = 2; i < 17; i++)
			if (addr->prv[i >> 1] & (0xf0 >> 4 * (i & 1)))
			    break;
		    while (i < 17) {
			value = (addr->prv[i >> 1] >> 4 * (1 - (i & 1))) & 0xf;
			if (value > 9)
			    return FATAL;
			if (!length--)
			    return FATAL;
			*buffer++ = '0' + value;
			i++;
		    }
		    if (!length--)
			return FATAL;
		    *buffer++ = ':';
		    i = 9;
		    fmt = e164;
		    break;
		default:
		    break;
	    }
	for (left = *fmt++; i < ATM_ESA_LEN; i++) {
	    if (!left--) {
		if (!length--)
		    return FATAL;
		*buffer++ = '.';
		left = *fmt++ - 1;
	    }
	    if (length < 2)
		return FATAL;
	    sprintf(buffer, "%02X", addr->prv[i]);
	    length -= 2;
	    buffer += 2;
	}
    }
    if (!length)
	return FATAL;
    *buffer = 0;
    return orig_len - length;
}


static int search_pvc(FILE *file, char *buffer, int length,
		      const struct sockaddr_natm *addr, int flags)
{
    struct sockaddr_natm temp;
    char line[MAX_ATM_NAME_LEN + 1];
    const char *here;
 
    while (fgets(line, MAX_ATM_NAME_LEN, file)) {
        if (!(here = strtok(line, "\t\n "))) 
	    continue;
	if (text2atm_pvc(here, &temp, sizeof(temp), flags) < 0)
	    continue;
	if (temp.snatm_family == AF_NATM) {
	    if ((addr->snatm_vpi != temp.snatm_vpi) ||
		(addr->snatm_vci != temp.snatm_vci))
	      continue;
	}
	while ((here = strtok(NULL, "\t\n ")))
	    if (strlen(here) < length) {
		strcpy(buffer, here);
		return 0;
	    }
	return FATAL;
    }
    return TRY_OTHER;
}

static int search_svc(FILE *file, char *buffer, int length,
		      const struct svc_addr *addr, int flags)
{
    struct svc_addr temp;
    char line[MAX_ATM_NAME_LEN + 1];
    const char *here;
 
    while (fgets(line, MAX_ATM_NAME_LEN, file)) {
        if (!(here = strtok(line, "\t\n "))) 
	    continue;
	if (text2atm_svc(here, &temp, sizeof(temp), flags) < 0)
	    continue;
	if (!atm_equal(addr, &temp, 0, 0))
	    continue;
        while ((here = strtok(NULL, "\t\n ")))
	    if (strlen(here) < length) {
		strcpy(buffer, here);
		return 0;
	    }
	return FATAL;
    }
    return TRY_OTHER;
}
 
 
static int try_name_pvc(char *buffer, int length, 
			const struct sockaddr_natm *addr)
{
    FILE *file;
    int result;
 
    if (!(file = fopen(HOSTS_ATM, "r")))
	return TRY_OTHER;
    result = search_pvc(file, buffer, length, addr, T2A_PVC);
    (void) fclose(file);
    return result;
}

static int try_name_svc(char *buffer, int length, const struct svc_addr *addr)
{
    FILE *file;
    int result;
 
    if (!(file = fopen(HOSTS_ATM, "r")))
	return TRY_OTHER;
    result = search_svc(file, buffer, length, addr, T2A_SVC);
    (void) fclose(file);
    return result;
}


int atm2text_pvc(char *buffer, int length, const struct sockaddr_natm *addr, 
		 int flags)
{
    int result;

    if (!length)
	return -1;
    if (flags & A2T_NAME) {
	result = try_name_pvc(buffer, length, addr);
	if (result == FATAL) 
	  return FATAL;
	if (result != TRY_OTHER) 
	  return strlen(buffer);
    }
    return do_pvc(buffer, length, addr, flags);
}

int atm2text_svc(char *buffer, int length, const struct svc_addr *addr, 
		 int flags)
{
    int result;

    if (!length)
	return -1;
    if (flags & A2T_NAME) {
	result = try_name_svc(buffer, length, addr);
	if (result == TRY_OTHER && !(flags & A2T_LOCAL))
	    result = ans_byaddr(buffer, length, addr, flags);
	if (result == FATAL)
	    return FATAL;
	if (result != TRY_OTHER)
	    return strlen(buffer);
    }
    return do_svc(buffer, length, addr, flags);
}
