/* text2atm.c - Converts textual representation of ATM address to binary
		encoding */

/* Written 1995-1998 by Werner Almesberger, EPFL-LRC/ICA */


#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <stdio.h>
#include <ctype.h>

#include <string.h>
#include <limits.h>
#include "../linux-atm.h"
#include "atm.h"
#include "atmsap.h" 
#include "atmres.h" 


static int try_pvc(const char *text, struct sockaddr_natm *addr, int flags)
{
    int part[3];
    int i;
  
    part[0] = part[1] = part[2] = 0;
    i = 0;
    while (1) {
	if (!*text)
	    return FATAL; /* empty or ends with a dot */
	if (i == 3)
	    return TRY_OTHER; /* too long */
	if (isdigit(*text)) {
	    if (*text == '0' && isdigit(text[1]))
		return TRY_OTHER;
	    /* no leading zeroes */
	    do {
		if (part[i] > INT_MAX / 10) 
		    return TRY_OTHER; /* number too big */
		part[i] = part[i] * 10 + *text++ - '0';
	    }
	    while (isdigit(*text))
		/* continue */;
	    i++;
	    if (!*text)
		break;
	    if (*text++ != '.')
		return TRY_OTHER; /* non-PVC character */
	    continue;
	}
	if (*text == '*') {
	    if (!(flags & T2A_WILDCARD))
		return FATAL; /* not allowed */
	    part[i++] = ATM_VCI_ANY; /* all *_ANY have the same value = 0 */
	}
#ifdef nomore
	/* we don't handle unspecified values for the moment, see later */
	else {
	    if (*text != '?')
		return TRY_OTHER; /* invalid character */
	    if (!(flags & T2A_UNSPEC))
		return FATAL; /* not allowed */
	    part[i++] = ATM_VPI_UNSPEC; /* all *_UNSPEC have the same value */
	}
#endif
	if (!*++text)
	    break;
	if (*text++ != '.')
	    return FATAL; /* dot required */
    }
    if (i < 2)
	return TRY_OTHER; /* no dots */
    if (i == 2) {
	part[2] = part[1];
	part[1] = part[0];
	part[0] = 0; /* default interface */
    }
    if (part[1] > ATM_MAX_VPI || part[2] > ATM_MAX_VCI)
	return TRY_OTHER; /* too big */
    if (part[1] > (flags & T2A_NNI ? ATM_MAX_VPI_NNI : ATM_MAX_VPI))
	return TRY_OTHER; /* too big */
    addr->snatm_family = AF_NATM;
    addr->snatm_len = sizeof(struct sockaddr_natm);
    snprintf(addr->snatm_if, IFNAMSIZ, ENNAME, part[0]);
    addr->snatm_vpi = part[1];
    addr->snatm_vci = part[2];
    return 0;
}

static int do_try_nsap(const char *text, struct svc_addr *addr, int flags)
{
    const char *walk;
    int count, pos, dot;
    int offset, len;
    char value;

    count = dot = 0;
    for (walk = text; *walk; walk++)
	if (isdigit(*walk)) {
	    if (count++ == 15)
		break;
	    dot = 1;
	} else if (*text != '.')
	    break;
	else if (!dot)
	    return FATAL; /* two dots in a row */
	else
	    dot = 0;
    if (*walk != ':') {
	pos = 0;
	offset = 0;
    } else {
	if (!dot || *text == '0')
	    return FATAL;
	addr->prv[0] = ATM_AFI_E164;
	addr->prv[1] = 0;
	memset(addr->prv + 1, 0, 8);
	for (pos = 18 - count - 1; *text; text++)
	    if (*text != '.')
		if (*text == ':')
		    break;
		else {
		    if (pos & 1)
			addr->prv[pos >> 1] |= *text - '0';
		    else
			addr->prv[pos >> 1] = (*text - '0') << 4;
		    pos++;
		}
	addr->prv[8] |= 0xf;
	text++;
	pos++;
	offset = 72;
    }
    for (dot = 0; *text; text++)
	if (isxdigit(*text)) {
	    if (pos == ATM_ESA_LEN * 2)
		return TRY_OTHER; /* too long */
	    value = isdigit(*text) ? *text - '0' :
			(islower(*text) ? toupper(*text) : *text) - 'A' + 10;
	    if (pos & 1)
		addr->prv[pos >> 1] |= value;
	    else
		addr->prv[pos >> 1] = value << 4;
	    pos++;
	    dot = 1;
	} else  if (*text == '/' && (flags & T2A_WILDCARD))
	    break;
	else if (*text != '.')
	    return TRY_OTHER;
	else {
	    if (!dot)
		return FATAL; /* two dots in a row */
	    dot = 0;
	}
    if (!dot)
	return FATAL;
    if (pos > 1 && !*addr->prv)
	return TRY_OTHER; /* no leading zeroes */
    if (!*text)
	return pos != ATM_ESA_LEN * 2 ? TRY_OTHER : ATM_ESA_LEN * 2;
    /* handle bad length */
    len = 0;
    while (*++text) {
	if (!isdigit(*text))
	    return -1; /* non-digit in length */
	if (len >= pos * 4)
	    return -1; /* too long */
	len = len * 10 + *text - '0';
    }
    if (len > 7 && addr->prv[0] != ATM_AFI_E164)
	offset = 72;
    if (len < offset)
	return FATAL;
    return len > pos * 4 ? TRY_OTHER : len;
}


static int try_nsap(const char *text, struct svc_addr *addr, int flags)
{
    int result;

    result = do_try_nsap(text, addr, flags);
    if (result < 0) 
	return result;
    *addr->pub = 0;
    return result;
}


static int try_e164(const char *text, struct svc_addr *addr, int flags)
{
    int i, dot, result;

    if (*text == ':' || *text == '+')
	text++;
    for (i = dot = 0; *text; text++)
	if (isdigit(*text)) {
	    if (i == ATM_E164_LEN)
		return TRY_OTHER; /* too long */
	    addr->pub[i++] = *text;
	    dot = 1;
	} else if (*text != '.')
	    break;
	else {
	    if (!dot)
		return TRY_OTHER; /* two dots in a row */
	    dot = 0;
	}
    if (!dot)
	return TRY_OTHER;
    addr->pub[i] = 0;
    *addr->prv = 0;
    result = 0;
    if (*text)
	if (*text++ != '+')
	    return TRY_OTHER;
	else {
	    result = do_try_nsap(text, addr, flags);
	    if (result < 0)
		return FATAL;
	}
    return result;
}


static int search_pvc(FILE *file, const char *text, 
		      struct sockaddr_natm *addr, int length,  int flags)
{
    char line[MAX_ATM_NAME_LEN + 1];
    const char *here;
    int result;

    while (fgets(line, MAX_ATM_NAME_LEN, file)) {
	if (!strtok(line, "\t\n ")) 
	    continue;
	while ((here = strtok(NULL, "\t\n ")))
	    if (!strcasecmp(here, text)) {
		here = strtok(line, "\t\n ");
		result = text2atm_pvc(here, addr, length, flags);
		if (result >= 0) 
		    return result;
	    }
    }
    return TRY_OTHER;
}

static int search_svc(FILE *file, const char *text, struct svc_addr *addr, 
		      int length, int flags)
{
    char line[MAX_ATM_NAME_LEN + 1];
    const char *here;
    int result;

    while (fgets(line, MAX_ATM_NAME_LEN, file)) {
	if (!strtok(line, "\t\n ")) 
	    continue;
	while ((here = strtok(NULL, "\t\n ")))
	    if (!strcasecmp(here, text)) {
		here = strtok(line, "\t\n ");
		result = text2atm_svc(here, addr, length, flags);
		if (result >= 0)
		    return result;
	    }
    }
    return TRY_OTHER;
}

static int try_name_pvc(char *buffer, struct sockaddr_natm *addr, int length, 
			int flags)
{
    FILE *file;
    int result;
 
    if (!(file = fopen(HOSTS_ATM, "r")))
	return TRY_OTHER;
    result = search_pvc(file, buffer, addr, length, flags);
    (void) fclose(file);
    return result;
}

static int try_name_svc(char *buffer, struct svc_addr *addr, int length, 
			int flags)
{
    FILE *file;
    int result;
 
    if (!(file = fopen(HOSTS_ATM, "r")))
	return TRY_OTHER;
    result = search_svc(file, buffer, addr, length, flags);
    (void) fclose(file);
    return result;
}


int text2atm_pvc(const char *text, struct sockaddr_natm *addr, int length,
		 int flags)
{
    int result;

    if (!*text)
	return -1;
    if (!(flags & T2A_PVC))
	flags |= T2A_PVC;
    if (length < sizeof(struct sockaddr_natm))
	return -1;
    result = try_pvc(text, addr, flags);
    if (result != TRY_OTHER) 
	return result;
    if (!(flags & T2A_NAME))
	return -1;
    result = try_name_pvc(text, addr, length, flags & ~T2A_NAME);
    if (result != TRY_OTHER) 
	return result;
    return -1;
}

int text2atm_svc(const char *text, struct svc_addr *addr, int length, 
		 int flags)
{
    int result;

    if (!*text)
	return -1;
    if (!(flags & T2A_SVC)) 
	flags |= T2A_SVC;
    if (length < sizeof(struct svc_addr))
	return -1;
    result = try_nsap(text, addr, flags);
    if (result != TRY_OTHER) 
	return result;
    result = try_e164(text, addr, flags);
    if (result != TRY_OTHER) 
	return result;
    if (!(flags & T2A_NAME))
	return -1;
    result = try_name_svc(text, addr, length, flags & ~T2A_NAME);
    if (result == TRY_OTHER && !(flags & T2A_LOCAL))
	result = ans_byname(text, addr, length, flags);
    if (result != TRY_OTHER)
	return result;
    return -1;
}
