/* fromscl.c - Processing of messages coming from signaling client */
 
/* (former kernel.c) Written 1995-1998 by Werner Almesberger, EPFL-LRC/ICA */
/* Modified 1997-1998 by Mohsen.Souissi@inria.fr, INRIA */
 
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"
#include "uni.h"
#include <q.out.h>
#include "qlib.h"

#include "proto.h"
#include "sap.h"
#include "io.h"
#include "policy.h"
#include "timeout.h"


#define COMPONENT "FROM_SCLIENT"

/* SETUP message is sent across the ATM network to */
/* establish a p2p or a p2m connection. */
static int send_setup(SOCKET *sock)
{
    static u_long call_ref = 0;
    struct svc_addr *local;
    SOCKET *walk;
    Q_DSC dsc;
    int error, size;

    do {
	if (++call_ref == 0x7fffff)
	    call_ref = 1;
	for (walk = sockets; walk; walk = walk->next)
	    if (walk->call_ref == call_ref)
		 break;
    } while (walk);
    sock->call_ref = call_ref;
    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_call_ref, call_ref);
    q_assign(&dsc, QF_msg_type, ATM_MSG_SETUP);
    q_assign(&dsc, QF_aal_type, 5); /* AAL 5 */
#ifdef UNI30
    /* Message mode - LANE seems to really want this */
    q_assign(&dsc, QF_aal_mode, 1);
#endif
    q_assign(&dsc, QF_sscs_type, 0); /* unspecified - LANE wants this */
    q_assign(&dsc, QF_bearer_class, 16); /* BCOB-X */
#if defined(UNI30) || defined(ALLOW_UNI30)
    /* force presence - UNI 3.0 wants this */
    q_assign(&dsc, QF_trans_cap, ATM_TC_VBR_NRT_R00);
#endif
    if (sock->conf == p2p) {
	q_assign(&dsc, QF_upcc, 0);	/* p2p */
	error = sap_encode(&dsc, &sock->remote, &sock->sap, &sock->qos);
    } else {			/* p2m case, root */
	q_assign(&dsc, QF_upcc, 1);
	q_assign(&dsc, QF_ep_ref, 0);
	q_assign(&dsc, QF_bw_pcr_01, 0);
	/* sock->parties points to the first and the only party */
	error = sap_encode(&dsc, &sock->parties->remote,
			   &sock->parties->sap, &sock->qos);
    }
#ifndef UNI30
    q_assign(&dsc, QF_qos_cs, Q2931_CS_ITU);
#endif
    q_assign(&dsc, QF_qos_fw, 0); /* QOS 0 */
    q_assign(&dsc, QF_qos_bw, 0);
    local = NULL;
    if (atmsvc_addr_in_use(sock->local))
	local = &sock->local;
    else if (local_addr[0].state == ls_same &&
	     atmsvc_addr_in_use(local_addr[0].addr))
	local = &local_addr[0].addr;
    if (!local) {
	if (!error)
	    error = -EADDRNOTAVAIL;
    } else if (*local->pub) {
	q_assign(&dsc, QF_cgpn_plan, ATM_NP_E164);
	q_assign(&dsc, QF_cgpn_type, ATM_TON_INTRNTNL);
	q_write(&dsc, QF_cgpn, (void *) local->pub, strlen(local->pub));
    } else {
	q_assign(&dsc, QF_cgpn_plan, ATM_NP_AEA);
	q_assign(&dsc, QF_cgpn_type, ATM_TON_UNKNOWN);
	q_write(&dsc, QF_cgpn, (void *) local->prv, ATM_ESA_LEN);
    }    
    if (net) {		/* see after... */
	if (!atmpvc_addr_in_use(sock->pvc)) {
	    int vci;

	    if (!allocate_ci)
		diag(COMPONENT, DIAG_FATAL, "No CI allocator (use -A)");
	    vci = get_vci(0);
	    if (vci < 0) {
		(void) q_close(&dsc);
		return vci;
	    }
	    strncpy(sock->pvc.snatm_if, get_itf(0), IFNAMSIZ);
	    sock->pvc.snatm_vpi = 0;
	    sock->pvc.snatm_vci = vci;
	}
	q_assign(&dsc, QF_vpi, sock->pvc.snatm_vpi);
	q_assign(&dsc, QF_vci, sock->pvc.snatm_vci);
    }
    if ((size = q_close(&dsc)) < 0) 
	error = -EINVAL;
    else if (!error) 
	to_signaling(q_buffer, size);
    return error;
}

static int send_add_party(SOCKET *sock, PARTY *party)
{
    Q_DSC dsc;
    int error, size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_call_ref, sock->call_ref);
    q_assign(&dsc, QF_msg_type, ATM_MSG_ADD_PARTY);
    q_assign(&dsc, QF_aal_type, 5); /* AAL 5 */
    error = sap_encode(&dsc, &party->remote, &party->sap, NULL);
    q_assign(&dsc, QF_ep_ref, party->ep_ref);
    /* Calling P. Number 5.4.5.13 */
    if (*sock->local.pub) {
	q_assign(&dsc, QF_cgpn_plan, ATM_NP_E164);
	q_assign(&dsc, QF_cgpn_type, ATM_TON_INTRNTNL);
	q_write(&dsc, QF_cgpn, (void *) sock->local.pub,
		strlen(sock->local.pub));
    } else {
	q_assign(&dsc, QF_cgpn_plan, ATM_NP_AEA);
	q_assign(&dsc, QF_cgpn_type, ATM_TON_UNKNOWN);
	q_write(&dsc, QF_cgpn, (void *) sock->local.prv, ATM_ESA_LEN);
    }
    if ((size = q_close(&dsc)) < 0) 
	error = -EINVAL;
    else if (!error) 
	to_signaling(q_buffer, size);
    return error;
}


static int send_connect(SOCKET *sock, PARTY *party)
{
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, ATM_MSG_CONNECT);
    q_assign(&dsc, QF_call_ref, sock->call_ref);
    q_assign(&dsc, QF_aal_type, 5);
    if (party != NULL)		/* leaf conf */
	q_assign(&dsc, QF_ep_ref, party->ep_ref | 0x8000);
    else {			/* p2p conf */
#ifdef UNI30
	/* Message mode - LANE seems to really want this */
	q_assign(&dsc, QF_aal_mode, 1);
#endif
	q_assign(&dsc, QF_sscs_type, 0); /* unspecified - LANE wants this */
	q_assign(&dsc, QF_fw_max_sdu, sock->qos.rxtp.max_sdu);
	q_assign(&dsc, QF_bw_max_sdu, sock->qos.txtp.max_sdu);
    }
    if ((size = q_close(&dsc)) >= 0) 
	to_signaling(q_buffer, size);
    return 0;
}

#ifdef Q2963_1

static void send_modify_request(SOCKET *sock)
{
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, ATM_MSG_MODIFY_REQ);
    q_assign(&dsc, QF_call_ref, sock->call_ref);
    if (sock->new_qos.txtp.traffic_class)
	q_assign(&dsc, QF_fw_pcr_01, SELECT_TOP_PCR(sock->new_qos.txtp));
    if (sock->new_qos.rxtp.traffic_class)
	q_assign(&dsc, QF_bw_pcr_01, SELECT_TOP_PCR(sock->new_qos.rxtp));
    if ((size = q_close(&dsc)) >= 0)
	to_signaling(q_buffer, size);
}

static void send_modify_ack(SOCKET *sock)
{
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, ATM_MSG_MODIFY_ACK);
    q_assign(&dsc, QF_call_ref, sock->call_ref);
    if (sock->qos.rxtp.traffic_class != ATM_NONE &&
	sock->qos.rxtp.max_pcr < sock->new_qos.rxtp.max_pcr) {
	q_assign(&dsc, QF_type_of_report, ATM_TOR_MOD_CONF);
        START_TIMER(sock, T361);
    }
    if ((size = q_close(&dsc)) >= 0)
	to_signaling(q_buffer, size);
}

#endif


static int bind_check(SOCKET *sock, struct svc_addr *local)   /* see after.. */
{
    struct svc_addr *use;

    if (atmsvc_addr_in_use(sock->local))
	return 0;
    if (atmsvc_addr_in_use(*local))
	use = local;
    else {
	if (local_addr[0].state != ls_same) {
	    diag(COMPONENT, DIAG_ERROR, "no local address");
	    return -EADDRNOTAVAIL;
	}
	use = &local_addr[0].addr;
    }
    sock->local = *use;
    return 0;
}

static int process_connect_sdu(SCLIENT *scl, SOCKET *sock, 
			       struct atmsvc_msg *msg)
{
    int error;
    CONN_CONF conf;

    if (sock) 
	return 1;
    conf = (msg->ep_ref < 0) ? p2p : root;
    if (!allow(&msg->svc, ACL_OUT)) {
	SEND_ERROR(scl, msg->vcc, -1, -EACCES);
	return -1;
    }
    sock = new_sock(scl, msg->vcc, conf);
    error = bind_check(sock, &msg->local);
    if (error) {
	SEND_ERROR(scl, msg->vcc, -1, error);
	free_sock(sock);
	return -1;
    }
    switch (sock->conf) {
      case p2p:
	sock->remote = msg->svc;
	sock->sap = msg->sap;
	break;
      case root:			/* root conf (p2m) */
	if (msg->ep_ref != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "ep_ref must be 0 (zero) for the first party");
	    error = -EINVAL;
	    SEND_ERROR(scl, msg->vcc, -1, error);
	    free_sock(sock);
	    return error;
	} else {
	    PARTY *first_p;
      
	    first_p = new_party(sock, 0, ps_add_init);
	    first_p->remote = msg->svc;
	    first_p->sap = msg->sap;
	    sock->nbactp = 1;
	}
	break;
      default:
	diag(COMPONENT, DIAG_ERROR, 
	     "process_connect_sdu: invalid ATM connection configuration");
	error = -EINVAL;
	SEND_ERROR(scl, msg->vcc, -1, error);
	free_sock(sock);
	return error;
    } 
    sock->qos = msg->qos;
    sock->state = ss_connecting;
    error = send_setup(sock);
    if (error) {
	SEND_ERROR(scl, msg->vcc, -1, error);
	free_sock(sock);
	return error;
    }
    START_TIMER(sock, T303);
    new_state(sock, ss_connecting);
    if (sock->conf == root) {
	sock->parties->pstate = ps_add_init;
	sock->nbactp++;
    }
    return 0;
}

static int process_add_sdu(SCLIENT *scl, SOCKET *sock, struct atmsvc_msg *msg)
{
    int error;
    u_short ep_ref;
    PARTY *walk, *party;

    if (sock == NULL) {
	diag(COMPONENT, DIAG_WARN, "process_add_sdu: no matching SOCKET");
	SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
	return 0;
    }
    if (sock->conf == p2p) { 
	diag(COMPONENT, DIAG_WARN, 
	     "process_add_sdu: attempt to add a party to a p2p connection");
	SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
	return 0;
    }
    if (sock->state != ss_connected) { 
	diag(COMPONENT, DIAG_WARN, 
	     "process_add_sdu: SOCKET is not in connected link state");
	SEND_ERROR(scl, msg->vcc, msg->ep_ref, -ENOTCONN);
	return 0;
    }
    if (sock->state == ss_rel_req)
	return 0;		/* ignore ADD_PARTY message (Note 5.6.3.3) */
    if (msg->ep_ref <= 0) {
	diag(COMPONENT, DIAG_WARN,
	     "process_add_sdu: %d: bad ep_ref, must be > 0", msg->ep_ref);
	SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
	return 0;
    }
    ep_ref = (u_short) msg->ep_ref;
    for (walk = sock->parties; walk != NULL; walk = walk->next)
	if (walk->ep_ref == ep_ref) {
	    diag(COMPONENT, DIAG_WARN,
		 "process_add_sdu: %d: ep_ref already exists");
	    SEND_ERROR(scl, msg->vcc, ep_ref, -EADDRINUSE);
	    return 0;
	} 
    if ((party = new_party(sock, ep_ref, ps_add_init)) == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_add_sdu: failed to allocate PARTY struct");
	SEND_ERROR(scl, msg->vcc, msg->ep_ref, -ENOMEM);
	return 0;
    }    
    party->remote = msg->svc;
    party->sap = msg->sap;
    error = send_add_party(sock, party);
    if (error != 0) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_add_sdu: failed to send add_party message");
	SEND_ERROR(scl, msg->vcc, msg->ep_ref, error);
	free_party(party);
	return 0;
    }
    START_TIMER(party, T399);
    sock->nbactp++;
    return 0;
}

static int process_accept_sdu(SCLIENT *scl, SOCKET *sock, PARTY *party, 
			      struct atmsvc_msg *msg)
{
    int error;

    if (sock->state == ss_zombie) {
	SEND_ERROR(scl, msg->vcc, -1, -ECONNABORTED);	/* -ERESTARTSYS ? */
	free_sock(sock);
	return 0;
    }
    if (sock->state != ss_indicated && sock->state != ss_proceeding)
	return 1;
    if (sock->state == ss_indicated && net)
	diag(COMPONENT, DIAG_FATAL, "No CI allocator (use -A) ???");
    if ((sock->conf == leaf) && (party->pstate != ps_add_recv))
	return 1;
    error = send_connect(sock, party);
    if (!error) {
	if (sock->conf == p2p)	/* no timers when conf == leaf */
	    START_TIMER(sock, T313);
	new_state(sock, ss_accepting);
	return 0;
    }
    SEND_ERROR(scl, sock->num, -1, error);
    send_release(sock->call_ref, 0); /* @@@ */
    START_TIMER(sock, T308_1);
    new_state(sock, ss_wait_rel);
    return 0;
}

static int process_reject_sdu(SOCKET *sock)
{
    switch (sock->state) {
      case ss_indicated:
	send_release_complete(sock->call_ref, ATM_CV_CALL_REJ);
	/* fall through */
      case ss_zombie:
	free_sock(sock);
	return 0;
      case ss_proceeding:
	/* @@ should use msg->reply */
	send_release(sock->call_ref, ATM_CV_CALL_REJ);
	START_TIMER(sock, T308_1);
	new_state(sock, ss_wait_rel);
	return 0;
      default:
	return 1;
    }
}

static int process_listen_sdu(SCLIENT *scl, SOCKET *sock, 
			      struct atmsvc_msg *msg)
{
    if (sock) 
	return 1;
    if (msg->qos.aal != ATM_AAL5) {
	SEND_ERROR(scl, msg->vcc, -1, -EINVAL);
	return 0;
    }
    if (lookup_sap(&msg->svc, &msg->sap, &msg->qos, NULL, NULL, NULL, 1)) {
	SEND_ERROR(scl, msg->vcc, -1, -EADDRINUSE);
	return 0;
    }
    sock = new_sock(scl, msg->vcc, p2p); /* by default conf = p2p */
    sock->local = msg->svc;
    sock->sap = msg->sap;
    sock->qos = msg->qos;
    send_sclient(scl, sock->num, 0, -1, as_okay,
		 0, NULL, NULL, NULL, NULL, NULL);
    sock->state = ss_listening;
    return 0;
}

static int process_close_sdu(SOCKET *sock)
{
    if (sock && ((sock->state == ss_indicated) || 
		 (sock->state == ss_proceeding) ||
		 (sock->state == ss_zombie) || 
		 (sock->state == ss_wait_rel))) 
	return 1;
    switch (sock ? sock->state : ss_null) {
      case ss_listening:
	send_close_sdu(sock);
	if (sock->listen)
	    new_state(sock, ss_listen_zombie);
	else
	    free_sock(sock);
	return 0;
      case ss_zombie:
	send_close_sdu(sock);
	/* fall through */
      case ss_wait_close:
	free_sock(sock);
	/* fall through */
      case ss_null:
      case ss_rel_req:
	return 0;
      case ss_connecting:
      case ss_accepting:
#ifdef Q2963_1
      case ss_mod_req:
#endif
	STOP_TIMER(sock);
	/* fall through */
#ifdef Q2963_1
      case ss_mod_lcl:
      case ss_mod_rcv:
      case ss_mod_fin_ok:
      case ss_mod_fin_fail:
      case ss_mod_fin_ack:
#endif
      case ss_connected:
#ifdef Q2963_1
	if (timer_handler(sock->conn_timer) == on_T361)
	    STOP_TIMER(sock);
#endif
	if (sock->state == ss_connected)
	    diag(COMPONENT, DIAG_INFO, "Active close <CR = 0x%06X>",
		 sock->call_ref);
	send_release(sock->call_ref,
#if defined(UNI31) || defined(UNI40)
		     ATM_CV_NORMAL_CLEAR
#else
		     ATM_CV_NORMAL_UNSPEC
#endif
		     );
	START_TIMER(sock, T308_1);
	new_state(sock, ss_rel_req);
	return 0;
      case ss_rel_ind:
	send_release_complete(sock->call_ref, 0); /* @@@ */
	free_sock(sock);
	return 0;
      default:
	return 1;
    }
}

static int process_drop_sdu(SCLIENT *scl, SOCKET *sock, struct atmsvc_msg *msg)
{
    PARTY *party;

    if (sock == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_drop_sdu: No matchig SOCKET for DROP message"); 
	SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
	return 0;
    }
    if (sock->parties == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_drop_sdu: SOCKET [%d, %ld] has no parties", 
	     sock->sclient->id, sock->num);
	SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL); 
	return 0;
    }
    for (party = sock->parties; party != NULL; party = party->next)
	if (party->ep_ref == (u_short) msg->ep_ref)
	    break;
    if (party == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_drop_sdu: No matchig party for "
	     "[scl_id, sock_num, ep_ref] = [%d, %ld, %d]",
	     sock->sclient->id, sock->num, msg->ep_ref); 
	SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
	return 0;
    }
    if (sock->nbactp == 1) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_drop_sdu: use <close> SDU instead (last active party)");
	SEND_ERROR(scl, msg->vcc, msg->ep_ref, -EINVAL);
	return 0;
    }
    switch (party->pstate) {
      case ps_add_init:
	STOP_TIMER(party);
	/* fall through */
      case ps_active:
	diag(COMPONENT, DIAG_INFO, "Active DROP <CR, EPR> = <%0x%06X, %d>",
	     sock->call_ref, party->ep_ref);
	send_drop_party(sock->call_ref, party->ep_ref, 
#if defined(UNI31) || defined(UNI40)
			ATM_CV_NORMAL_CLEAR
#else
			ATM_CV_NORMAL_UNSPEC
#endif
			);
	START_TIMER(party, T398);
	party->pstate = ps_drop_init;
	sock->nbactp--;
	return 0;
      case ps_drop_recv:
	diag(COMPONENT, DIAG_DEBUG, 
	     "Passive DROP <CR, EPR> = <%0x%06X, %d> complete",
	     sock->call_ref, party->ep_ref);
	free_party(party);
	break;
      default:
	break;
    }
    return 0;
}

static int process_identify_sdu(SCLIENT *scl, SOCKET *sock, PARTY *party,
				struct atmsvc_msg *msg)
{
    if (sock->state != ss_indicated && sock->state != ss_proceeding)
	return 1;
    if (!atmpvc_addr_in_use(msg->pvc)) {
	if (net && !allocate_ci)
	    diag(COMPONENT, DIAG_FATAL, "No CI allocator (use -A)");
	return 0;
    }
    if (net && allocate_ci)
	diag(COMPONENT, DIAG_FATAL, "CI allocation role conflict");
    sock->pvc = msg->pvc;
    if (send_call_proceeding(sock, party))
	diag(COMPONENT, DIAG_FATAL, "s_c_p failed");
    new_state(sock, ss_proceeding);
    return 0;
}

#ifdef Q2963_1

static int process_modify_sdu(SCLIENT *scl, SOCKET *sock,
			      struct atmsvc_msg *msg)
{
    if (sock && ((sock->state == ss_mod_lcl) ||
		 (sock->state == ss_mod_req) ||
		 (sock->state == ss_mod_rcv) ||
		 (sock->state == ss_mod_fin_ok) ||
		 (sock->state == ss_mod_fin_fail) ||
		 (sock->state == ss_mod_fin_ack))) {
	send_sclient(scl, sock->num, 0, -1, as_okay,
		     -EALREADY, NULL, NULL, NULL, NULL, NULL);
	return 0;
    }
    if (!sock || (sock->state != ss_connected)) {
	send_sclient(scl, sock->num, 0, -1, as_okay,
		     -EBADF, NULL, NULL, NULL, NULL, NULL);
	return 0;
    }
    if ((sock->qos.txtp.traffic_class != msg->qos.txtp.traffic_class) ||
	(sock->qos.rxtp.traffic_class != msg->qos.rxtp.traffic_class)) {
	/* @@@ may do more checking */
	send_sclient(scl, sock->num, 0, -1, as_okay,
		     -EINVAL, NULL, NULL, NULL, NULL, NULL);
	return 0;
    }
    sock->new_qos = msg->qos;
    send_sclient(scl, sock->num, 0, -1, as_modify,
		 ATM_MF_INC_RSV | ATM_MF_DEC_SHP,
		 NULL, NULL, NULL, NULL, &msg->qos);
    new_state(sock, ss_mod_lcl);
    return 0;
}

static int process_okay_sdu(SCLIENT *scl, SOCKET *sock, struct atmsvc_msg *msg)
{
    switch (sock ? sock->state : ss_null) {
      case ss_mod_lcl:
	send_modify_request(sock);
	START_TIMER(sock, T360);
	new_state(sock, ss_mod_req);
	return 0;
      case ss_mod_rcv:
	send_modify_ack(sock);
	sock->qos = sock->new_qos;
	new_state(sock, ss_connected);
	return 0;
      case ss_mod_fin_ok:
	send_sclient(scl, sock->num, 0, -1, as_okay,
		     0, NULL, NULL, NULL, NULL, NULL);
	new_state(sock, ss_connected);
	return 0;
      case ss_mod_fin_fail:
	send_sclient(scl, sock->num, 0, -1, as_okay,
		     sock->error, NULL, NULL, NULL, NULL, NULL);
	sock->error = 0;
	/* fall through */
      case ss_mod_fin_ack:
	new_state(sock, ss_connected);
	/* fall through */
      default:
	/* ignore stray as_okay */
	return 0;
    }
}

static int process_error_sdu(SCLIENT *scl, SOCKET *sock,
			     struct atmsvc_msg *msg)
{
    switch (sock ? sock->state : ss_null) {
      case ss_mod_lcl:
	send_sclient(scl, sock->num, 0, -1, as_okay,
		     msg->reply, NULL, NULL, NULL, NULL, NULL);
	new_state(sock, ss_connected);
	return 0;
      case ss_mod_rcv:
	send_modify_reject(sock->call_ref, ATM_CV_RES_UNAVAIL);
	new_state(sock, ss_connected);
	return 0;
      case ss_mod_fin_ok:
	diag(COMPONENT, DIAG_ERROR, "QOS commit failed");
	send_sclient(scl, sock->num, 0, -1, as_okay,
		     0, NULL, NULL, NULL, NULL, NULL);
	/* @@@ clear call instead ? */
	new_state(sock, ss_connected);
	return 0;
      case ss_mod_fin_fail:
	diag(COMPONENT, DIAG_ERROR, "QOS rollback failed");
	send_sclient(scl, sock->num, 0, -1, as_okay,
		     sock->error, NULL, NULL, NULL, NULL, NULL);
	sock->error = 0;
	/* @@@ clear call instead ? */
	new_state(sock, ss_connected);
	return 0;
      case ss_mod_fin_ack:
	/* @@@ maybe we should even clear the call now */
	diag(COMPONENT, DIAG_ERROR, "QOS commit failed");
	new_state(sock, ss_connected);
	return 0;
     default:
       /* ignore stray as_error */
       return 0;
    }
}

#endif

static void dispatch(SCLIENT *scl, SOCKET *sock, PARTY *party, 
		    struct atmsvc_msg *msg)
{
    int error;
  
    if (scl->state == DYING) {
	if (sock && sock->state != ss_wait_rel) {
	    send_release(sock->call_ref, 0);	/* @@@ */
						/* see reason after... */
	    START_TIMER(sock, T308_1);
	    new_state(sock, ss_wait_rel);
	}
	diag(COMPONENT, DIAG_DEBUG, 
	     "DYING sclient (%d), receiving message (%d)", 
	     scl->id, msg->type);
	return;
    }
    switch (msg->type) {

      /* only in NULL state */
      case as_bind:
	if (sock)
	    break;
	if (!atmsvc_addr_in_use(msg->svc))
#ifdef BE_PICKY_ABOUT_BINDING_LOCAL_WILDCARD_ADDRESSES
	    if (local_addr[0].state != ls_same)
		SEND_ERROR(scl, msg->vcc, -1, -EADDRNOTAVAIL);
	    else
#endif
	    send_sclient(scl, msg->vcc, 0, -1, as_okay,
			 0, NULL, NULL, &local_addr[0].addr, NULL, NULL);
	else {
		LOCAL_ADDR *walk;

		for (walk = local_addr; walk->state != ls_unused; walk++)
		    if (walk->state == ls_same &&
			atm_equal(&walk->addr, &msg->svc,
				  (ATM_ESA_LEN - 1) * 8, AXE_WILDCARD))
			break;
		    if (*msg->svc.pub)
			diag(COMPONENT, DIAG_DEBUG,
			     "binding to E.164 address %s\n", msg->svc.pub);
		    if (walk->state == ls_unused)
			SEND_ERROR(scl, msg->vcc, -1, -EADDRNOTAVAIL);
		    else
			send_sclient(scl, msg->vcc, 0, -1, as_okay,
				     0, NULL, NULL, NULL, NULL, NULL);
	}
	return;

      /* NULL state only. Used also for the first party of p2m */
      case as_connect:
	error = process_connect_sdu(scl, sock, msg);
	if (error == 1)
	    break;
	return;

      /* root (p2m) conf. For the first party, use as_connect */
      case as_add:
	error = process_add_sdu(scl, sock, msg);
	if (error == 1)
	    break;
	return;

      /* sock exists */
      case as_accept:
	error = process_accept_sdu(scl, sock, party, msg);
	if (error == 1)
	    break;
	return;

      /* ZOMBIE, INDICATED, or PROCEEDING */
      case as_reject:
	error = process_reject_sdu(sock);
	if (error == 1)
	    break;
	return;

      /* NULL */
      case as_listen:
	error = process_listen_sdu(scl, sock, msg);
	if (error == 1)
	    break;
	return;

      /* all but INDICATED, PROCEEDING, ZOMBIE, and WAIT_REL */
      case as_close:
	error = process_close_sdu(sock);
	if (error == 1)
	    break;
	return;

      /* for the last party, use as_close */
      case as_drop:
	error = process_drop_sdu(scl, sock, msg);
	if (error == 1)
	    break;
	return;

      case as_identify:
	error = process_identify_sdu(scl, sock, party, msg);
	if (error == 1)
	    break;
	return;

#ifdef Q2963_1
      case as_modify:
	error = process_modify_sdu(scl, sock, msg);
	if (error == 1)
	    break;
	return;

      case as_okay:
	error = process_okay_sdu(scl, sock, msg);
	if (error == 1)
	    break;
	return;

      case as_error:
	error = process_error_sdu(scl, sock, msg);
	if (error == 1)
	    break;
	return;
#endif

      /* fall through */
      default:
	diag(COMPONENT, DIAG_WARN, "invalid message %d", (int) msg->type);
	return;
    }
    diag(COMPONENT, DIAG_WARN, "message %s is incompatible with state %s (%d)",
	 as_name[msg->type], state_name[sock ? sock->state : ss_null],
	 (int) (sock ? sock->state : ss_null));
    return;
}

/* msg->type = as_accept || as_reject, maybe there's a party */
static void dispatch_listen(SCLIENT *scl, SOCKET *sock, PARTY *party, 
			    struct atmsvc_msg *msg)
{
    SOCKET *next;
    PARTY *currp = NULL;
    short ep_ref;

    if (!sock) {
	diag(COMPONENT, DIAG_WARN, "message %s is incompatible with state %s "
	     "(%d)", as_name[msg->type], state_name[ss_null], ss_null);
	SEND_ERROR(scl, msg->vcc, -1, -EINVAL);
	return;
    }
    if (!(next = sock->listen)) {
	diag(COMPONENT, DIAG_WARN,
	     "socket [%d, %ld]  got accept with empty listen queue", 
	     scl->id, msg->vcc);
	return;
    }
    sock->listen = next->listen;
    if (sock->state == ss_listen_zombie && !sock->listen)
	free_sock(sock);
    next->listen = NULL;
    if (next->state == ss_zombie) {
	SEND_ERROR(scl, next->num, -1, -ECONNABORTED);
	diag(COMPONENT, DIAG_WARN, "Passive connection aborted");
	free_sock(next);
	return;
    }
    if ((ep_ref = msg->ep_ref) >= 0) { /* it must be a leaf */
	if (next->conf != leaf) {
	    diag(COMPONENT, DIAG_WARN, 
		 "Bad SOCKET/message configuration/fields for sock [%d, %ld] "
		 "<EPR = %d>", scl->id, next->num, ep_ref);
	    SEND_ERROR(scl, next->num, -1, -EINVAL);
	    return;
	}
	for (currp = next->parties; currp != NULL; currp = currp->next)
	    if (currp->ep_ref == ep_ref)
		break;
	if (currp == NULL) {
	    diag(COMPONENT, DIAG_WARN, 
		 "No matchig party for [scl_id, sock_num, ep_ref] = "
		 "[%d, %ld, %d]", scl->id, next->num, ep_ref);
	    SEND_ERROR(scl, next->num, -1, -EINVAL);
	    return;
	}
    }
    dispatch(scl, next, currp, msg);
}


void itf_load(char *ifname)
{
    char buf[MAX_ATM_ADDR_LEN + 1];
    LOCAL_ADDR *walk;
    int itf;

    if (sscanf(ifname, ENNAME, &itf) != 1)
	diag(COMPONENT, DIAG_FATAL, "itf_load doesn't understand %s", ifname);
    (void) get_addr(itf);
    for (walk = local_addr; walk->state != ls_unused; walk++)
	if (walk->itf == itf) {
	    if (atm2text_svc(buf, MAX_ATM_ADDR_LEN + 1,
			     &walk->addr, pretty) < 0)
		strcpy(buf,"<invalid>");
	    switch (walk->state) {
		case ls_added:
		    diag(COMPONENT, DIAG_INFO,
			 "Added local ATM address %s at itf %d", buf, itf);
		    walk->state = ls_same;
		    break;
		case ls_removed:
		    diag(COMPONENT, DIAG_INFO,
			 "Removed local ATM address %s at itf %d", buf, itf);
		    walk->state = ls_unused;
		    /* @@@ delete SVCs using that address ? */
		    break;
		default:
		    break;
	    }
	}
}


void from_sclient(SCLIENT *scl, struct atmsvc_msg *msg, int size)
{
    void (*dispatcher)(SCLIENT *, SOCKET *, PARTY *, struct atmsvc_msg *);
    SOCKET *curr;
    PARTY *party = NULL;
    short ep_ref; 
  
    if (msg->type == as_terminate) {
	if (mode != sm_switch) {
	    diag(COMPONENT, DIAG_ERROR,
		 "Ignoring as_terminate received in %s mode",
		 mode == sm_user ? "USER" : "NETWORK");
	    return;
	}
	clear_all_calls();
	stop = 1;
	diag(COMPONENT, DIAG_INFO, "Going down on as_terminate");
	return;
    }
    if (msg->listen_vcc && ((msg->type == as_accept) ||
			    (msg->type == as_reject) ||
			    (msg->type == as_identify))) {
	dispatcher = dispatch_listen;
	for (curr = sockets; curr; curr = curr->next)
	    if ((curr->sclient == scl) && 
		(curr->num == msg->vcc) && 
		((curr->state == ss_listening) ||
		 (curr->state == ss_listen_zombie)))
		break;
    } else {
	dispatcher = dispatch;
	for (curr = sockets; curr; curr = curr->next)
	    if ((curr->sclient == scl) && (curr->num == msg->vcc)) 
		break; 
	if (curr != NULL)
	    if ((ep_ref = msg->ep_ref) >= 0)
		for (party = curr->parties; party != NULL; party = party->next)
		    if (party->ep_ref == ep_ref)
			break;
    }
    if (party != NULL) 
	diag(COMPONENT, DIAG_DEBUG,
	     "%s for party [%d, %ld, %d], [vcc/listen_vcc] = "
	     "[%ld/%ld], in sock/party state %s/%s",
	     as_name[msg->type], scl->id, curr->num, party->ep_ref, 
	     msg->vcc, msg->listen_vcc, state_name[curr->state], 
	     ps_name[party->pstate]);
    else
	diag(COMPONENT, DIAG_DEBUG,
	     "%s for socket [%d, %ld], [vcc/listen_vcc] = "
	     "[%ld/%ld], in state %s",
	     as_name[msg->type], scl->id, curr ? curr->num : 0, msg->vcc, 
	     msg->listen_vcc, state_name[curr ? curr->state : ss_null]);
    dispatcher(scl, curr, party, msg);
}
