/* io.c - I/O operations */
 
/* Written 1995-1998 by Werner Almesberger, EPFL-LRC/ICA */
 

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_atm.h>
#include <net/if_dl.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netnatm/natm.h>

#include <sys/time.h>
#include <sys/types.h>
#include <netdb.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"
#include "localnsap.h"
#include "uni.h"
#include "pdu.h"

#include "proto.h"
#include "io.h"
#include "trace.h"


#define COMPONENT "IO"
#define BACKLOG 5

struct timeval now;
struct sockaddr_natm signaling_pvc;
extern const struct in6_addr in6addr_any;

int stop = 0;

static int sclient_listen = -1, signaling;


/* ----- sclient interface ----------------------------------------------- */

/* open a TCP socket to communicate with signalling clients */
static int open_sclient(void)   
{
    int s, on = 1;
    struct servent *servp;
    struct sockaddr_in6 local_in6;

    if ((servp = getservbyname("atmsig", "tcp")) == 0) {
	perror("getservbyname");
	return (-1);
    }
    if ((s = socket(PF_INET6, SOCK_STREAM, 0)) < 0) {
	perror("socket");
	return (-1);
    }
    memset(&local_in6, 0, sizeof(local_in6));
    local_in6.sin6_len = sizeof(local_in6);
    local_in6.sin6_family = AF_INET6;
    local_in6.sin6_addr = in6addr_any;
    local_in6.sin6_port =  servp->s_port;
    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on)) < 0)
	perror("setsockopt");
    if (bind(s, (struct sockaddr *) &local_in6, sizeof(local_in6)) < 0) {
	perror("bind");
	return (-1);
    }
    if (listen(s, BACKLOG) < 0) {
	perror("listen");
	return (-1);
    }
    
    return (s);
}


static void recv_sclient(SCLIENT *scl)
{
    static long abuffer[sizeof(struct atmsvc_msg) / sizeof(long) + 1];
    u_char *buffer = (unsigned char *)abuffer;
    int size;

    /* now messages have a fixed size ! */
    size = read(scl->id, buffer, sizeof(struct atmsvc_msg));
    if (size < 0) {
	diag(COMPONENT, DIAG_ERROR, "recv sclient: %s", strerror(errno));
	return;	
    }
    if (size == 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "recv sclient: tcp_socket (%d) closed by peer", scl->id);
	scl->state = DYING;
	free_sclient(scl); /* sclient's connections will be torn down */ 
	return;
    }
    if (size < sizeof(struct atmsvc_msg))
	diag(COMPONENT, DIAG_FATAL, "sclient message too short (%d < %d)",
	     size, sizeof(struct atmsvc_msg));
    trace_sclient("FROM SCLIENT", scl->id, (struct atmsvc_msg *) buffer);
    from_sclient(scl, (struct atmsvc_msg *) buffer, size);
}


void to_sclient(SCLIENT *scl, struct atmsvc_msg *msg)
{
    int wrote;
   
    diag("SCLIENT", DIAG_DEBUG,
	 "TO SCLIENT: %s <reply = %d> for socket [%d, %ld], "
	 "<listen_vcc = %ld>, <EPR = %d>, <length =%d>", 
	 as_name[msg->type], msg->reply, scl->id, msg->vcc, 
	 msg->listen_vcc, msg->ep_ref, sizeof(*msg));
    
    /* should be "IO" ... */
    trace_sclient("TO SCLIENT", scl->id, msg);
    wrote = write(scl->id, msg, sizeof(*msg));
    if (wrote == sizeof(*msg))
	return;
    if (wrote < 0) {
	fprintf(stderr,
		"sclient %d write: %s",
		scl->id,
		strerror(errno));
	if (errno == EPIPE)
	    {
		diag(COMPONENT, DIAG_ERROR, "sclient %d dying", scl->id);
		scl->state = DYING; 
		/* sclient's connections will be torn down */ 
		free_sclient(scl); /* see this after... */
	    }
	return;
    }
    diag(COMPONENT, DIAG_ERROR, "bad sclient %d write: wanted %d, wrote %d",
	 scl->id, sizeof(*msg), wrote);
}


static void close_sclient(void)
{
    (void) close(sclient_listen);
}


/* ----- signaling interface ----------------------------------------------- */


static int open_signaling(void)
{
    struct atm_qos qos;
    int s;

    if ((s = socket(PF_NATM, SOCK_STREAM, PROTO_NATMAAL5)) < 0) {
	perror("socket");
	return (-1);
    }
    memset(&qos, 0, sizeof(qos));
    qos.aal = ATM_AAL5;
    qos.rxtp.max_sdu = qos.txtp.max_sdu = MAX_Q_MSG;
    if (sig_qos) {
	if (text2qos(sig_qos, &qos, T2Q_DEFAULTS) < 0) {
	    fprintf(stderr, "invalid qos: %s\n", sig_qos);
	    return (-1);
	}
    } else {
	if (sig_pcr == -1)
	    qos.rxtp.traffic_class = qos.txtp.traffic_class = ATM_UBR;
	else {
	    qos.rxtp.traffic_class = qos.txtp.traffic_class = ATM_CBR;
	    qos.rxtp.min_pcr = qos.txtp.min_pcr = sig_pcr;
	}
    }
    signaling_pvc.snatm_family = AF_NATM;
    if (connect(s,
		(struct sockaddr *) &signaling_pvc,
		sizeof(signaling_pvc)) < 0) {
	perror("connect");
	return (-1);
    }
    if (qos.txtp.max_pcr) {
	struct vctxreq vreq;

	memset(&vreq, 0, sizeof(vreq));
	strncpy(vreq.vc_ifname, signaling_pvc.snatm_if, IFNAMSIZ);
	vreq.vc_pcr = qos.txtp.max_pcr;
	if (ioctl(s, SIOCTXSPEED, &vreq) < 0) {
	    perror("ioctl SIOCTXSPEED");
	    return (-1);
	}
    }
    return (s);
}


static void recv_signaling(void)
{
    static long abuffer[MAX_Q_MSG / sizeof(long) + 1];
    unsigned char *buffer = (unsigned char *)abuffer;
    int size;

    size = read(signaling, buffer, MAX_Q_MSG);
    if (size < 1) {
	perror("read signaling");
	return;
    }
    diag(COMPONENT, DIAG_DEBUG, "FROM NET: %s PDU (%d bytes)",
	 pdu_name[size > 3 && !(size & 3) ? buffer[size - 4] & 0xf : 0],
	 size);
    if (debug)
	diag_dump(COMPONENT, DIAG_DEBUG, NULL, buffer, size);
    from_net(buffer, size);
}


void to_net(void *msg, int size)
{
    int wrote;

    diag(COMPONENT, DIAG_DEBUG, "TO NET: %s PDU (%d bytes)",
	 pdu_name[size > 3 && !(size & 3) ?
		    ((unsigned char *) msg)[size - 4] & 0xf : 0],
	 size);
    if (debug)
	diag_dump(COMPONENT, DIAG_DEBUG, NULL, msg, size);
    wrote = write(signaling, msg, size);
    if (wrote == size)
	return;
    if (wrote < 0) {
	perror("signaling write");
	return;
    }
    diag(COMPONENT, DIAG_WARN, "bad signaling write: wanted %d, wrote %d", 
	 size, wrote);
}


static void close_signaling(void)
{
    (void) close(signaling);
}

/* ----- addresses --------------------------------------------------------- */

LOCAL_ADDR local_addr[MAX_ADDRS + 1];


int get_addr(int itf)
{
    struct svc_addr buffer[MAX_ADDRS];
    LOCAL_ADDR *from, *to;
    int addrs = 0, i;

    for (from = to = local_addr; from->state != ls_unused; from++)
	if (from->state != ls_removed) {
	    from->state = itf == from->itf ? ls_removed : ls_same;
	    *to++ = *from;
	}
    bzero(buffer, sizeof(buffer));
    get_local_addr(buffer, MAX_ADDRS, itf);
    for (i = 0; i < MAX_ADDRS; i++) {
	if (atmsvc_addr_in_use(buffer[i]))
	    addrs++;
	else
	    continue;
	for (from = local_addr; from->state != ls_unused; from++)
	    if (from->itf == itf && atm_equal(buffer + i, &from->addr, 0, 0))
		break;
	if (from->state != ls_unused)
	    from->state = ls_same;
	else if (to == local_addr + MAX_ADDRS - 1)
	    diag(COMPONENT, DIAG_WARN, "local address table overflow");
	else {
	    to->state = ls_added;
	    to->itf = itf;
	    to->addr = buffer[i];
	    to++;
	}
    }
    to->state = ls_unused;
    return addrs;
}


/* ----- common part ------------------------------------------------------- */


int open_all(void)
{
    if (sclient_listen == -1)
	sclient_listen = open_sclient();
    if (sclient_listen < 0) 
	return (-1);
    signaling = open_signaling();
    if (signaling < 0) {
	close(sclient_listen);
	return (-1);
    }
    local_addr[0].state = ls_unused;
    return (0);
}


void close_all(void)    /* see later */
{
    SCLIENT *walk;
    SOCKET *brother;
    
    for (walk = sclients; walk; walk = walk->next)  {
	for (brother = walk->connections; brother; brother = brother->brother)
	    free_sock(brother);
	free_sclient(walk);
    }
    close_sclient();           
    close_signaling();
}


void init_current_time(void)
{
    (void) gettimeofday(&now, NULL);
}


void poll_loop(void)               
{
    fd_set perm, set;
    int fds, ret;
    SCLIENT *walk;
    int new_scl_id;
    struct sockaddr_in6 sclient_addr6;
    int len;

    FD_ZERO(&perm);
    FD_SET(sclient_listen, &perm);
    FD_SET(signaling, &perm);
    fds = FD_SETSIZE;
    (void) gettimeofday(&now, NULL);
    while (!stop) {
	set = perm;
	poll_signals();
	/*
	 * Here we have a small race condition: if a signal is delivered after
	 * poll_signals tests for it but before select sleeps, we miss that
	 * signal. If it is sent again, we're of course likely to get it. This
	 * isn't worth fixing, because those signals are only used for
	 * debugging anyway.
	 */
	 /* after... : special processing for DYING? */
	for (walk = sclients; walk; walk = walk->next) 
	    if (walk->state != DYING)
		FD_SET(walk->id, &set);
	ret = select(fds, &set, NULL, NULL, next_timer());
	if (ret < 0) {
	    if (errno != EINTR)
		perror("select");
	} else {
	    diag(COMPONENT, DIAG_DEBUG, "----------");
	    (void) gettimeofday(&now, NULL);
	    if (FD_ISSET(signaling, &set))
		recv_signaling();
	    for (walk = sclients; walk; walk = walk->next) 
		if (FD_ISSET(walk->id, &set))
		    recv_sclient(walk);
	    if (FD_ISSET(sclient_listen, &set)) {
		len = sizeof(struct sockaddr_in6);
		if ((new_scl_id = accept(sclient_listen, 
					 (struct sockaddr *) &sclient_addr6, 
					 &len)) < 0) 
		    perror("accept"); 
                else
		    new_sclient(new_scl_id);
		}
	    expire_timers();
	      /* expire timers after handling messges to make sure we don't
		 time out unnecessarily because of scheduling delays */
	}
    }
}


/*
 * The allocation strategy could be improved as follows: we should try
 * vci = prev_vci++  first and only resort to ATM_VCI_ANY if that fails several
 * times (and we should actually skip over those which are in use by SVCs. This
 * way we avoid using VCIs that just became available. Doing it "right" seems
 * to be getting complex, though.
 */


int get_pvc(int itf, int *vci)		/* see QoS after... */
{
    struct sockaddr_natm addr;
    struct atm_qos qos;
    int s, error;

    if ((s = socket(PF_NATM, SOCK_STREAM, PROTO_NATMAAL5)) < 0)
        diag(COMPONENT, DIAG_FATAL, "get_pvc: %s", strerror(errno));
    memset(&qos, 0, sizeof(qos));
    qos.aal = ATM_AAL5;
    qos.rxtp.traffic_class = qos.txtp.traffic_class = ATM_UBR;
    qos.rxtp.max_sdu = qos.txtp.max_sdu = 1; /* smallest possible SDU size */
#ifdef notyet
    if (setsockopt(s,SOL_ATM,SO_ATMQOS,&qos,sizeof(qos)) < 0)
	diag(COMPONENT,DIAG_FATAL,"setsockopt SO_ATMQOS: %s",strerror(errno));
#endif
    memset(&addr, 0, sizeof(addr));
    addr.snatm_family = AF_NATM;
    addr.snatm_len = sizeof(struct sockaddr_natm);
    snprintf(addr.snatm_if, IFNAMSIZ, ENNAME, itf);
    addr.snatm_vpi = 0; /* @@@ */
    addr.snatm_vci = ATM_VCI_ANY;
    error = 0;
    if (connect(s, (struct sockaddr *) &addr, sizeof(addr)) < 0)
	error = errno;
    else {
	int alen;

	alen = sizeof(addr);
	if (getpeername(s, (struct sockaddr *) &addr, &alen) < 0)
	    diag(COMPONENT, DIAG_FATAL, "get_pvc: %s", strerror(errno));
	*vci = addr.snatm_vci;
	return (s);
    }
    (void) close(s);
    return (-error);
}
