/* timeout.c - Processing of signaling timeout events */
 
/* Written 1995-1998 by Werner Almesberger, EPFL-LRC/ICA */
 
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include "../linux-atm.h"
#include "atmd.h"

#include "uni.h"
#include "proto.h"
#include "timeout.h"


#define COMPONENT "UNI"


static void complain(SOCKET *sock, const char *timer)
{
    diag(COMPONENT, DIAG_FATAL, "Timer %s expired in incompatible state %s",
	 timer, state_name[sock->state]);
}


void on_T303(void *user) /* CONNECTING */
{
    SOCKET *sock = user;

    diag(COMPONENT, DIAG_DEBUG, "T303 on sock [%d, %ld]",
	 sock->sclient->id, sock->num);
    if (sock->state != ss_connecting)
	complain(sock, "T303");
    SEND_ERROR(sock->sclient, sock->num, -1, -ETIMEDOUT);
    sock->conn_timer = NULL;
    free_sock(sock);
}


void on_T308_1(void *user) /* WAIT_REL or REL_REQ */
{
    SOCKET *sock = user;

    diag(COMPONENT,DIAG_DEBUG,"T308_1 on sock [%d, %ld]",
	 sock->sclient->id, sock->num);
    if (sock->state != ss_wait_rel && sock->state != ss_rel_req)
	complain(sock, "T308_1");
    send_release(sock->call_ref, ATM_CV_TIMER_EXP, 308); /* @@@ ? */
    sock->conn_timer = NULL;
    START_TIMER(sock, T308_2);
}


void on_T308_2(void *user) /* WAIT_REL or REL_REQ */
{
    SOCKET *sock = user;

    diag(COMPONENT, DIAG_WARN, "Trouble: T308_2 has expired");
    if (sock->state != ss_wait_rel && sock->state != ss_rel_req)
	complain(sock, "T308_2");
    sock->conn_timer = NULL;
    if (sock->state == ss_rel_req)
	send_close_sdu(sock);
    free_sock(sock);
}


void on_T309(void *user)
{
    diag(COMPONENT, DIAG_DEBUG, "T309 has expired");
    clear_all_calls_on_T309();
}


void on_T310(void *user)
{
    on_T303(user);
    diag(COMPONENT, DIAG_DEBUG, "(it's actually T310)");
}


void on_T313(void *user) /* ACCEPTING */
{
    SOCKET *sock = user;

    diag(COMPONENT, DIAG_DEBUG, "T313 on sock [%d, %ld]",
	 sock->sclient->id, sock->num);
    if (sock->state != ss_accepting)
	complain(sock, "T313");
    send_release(sock->call_ref, ATM_CV_TIMER_EXP, 313);
    sock->conn_timer = NULL;
    START_TIMER(sock, T308_1);
    new_state(sock, ss_rel_req);
}

#ifdef Q2963_1

void on_T360(void *user)
{
    SOCKET *sock = user;

    diag(COMPONENT, DIAG_DEBUG, "T360 on 0x%lx",
	 sock->sclient->id, sock->num);
    if (sock->state != ss_mod_req)
	complain(sock, "T360");
    send_release(sock->call_ref, ATM_CV_TIMER_EXP, 360);
    sock->conn_timer = NULL;
    START_TIMER(sock, T308_1);
    new_state(sock, ss_rel_req);
}

void on_T361(void *user)
{
    SOCKET *sock = user;

    diag(COMPONENT, DIAG_DEBUG, "T361 on 0x%lx",
	 sock->sclient->id, sock->num);
    if (sock->state != ss_connected)
	complain(sock, "T361");
    sock->qos = sock->new_qos;
    send_sclient(sock->sclient, sock->num, 0, -1, as_modify,
		 ATM_MF_SET, NULL, NULL, NULL, NULL, &sock->qos);
    sock->conn_timer = NULL;
    new_state(sock, ss_mod_fin_ack);
}

#endif

void on_T398(void *user) /* ps_DROP_INIT */
{
    PARTY *party = user;

    diag(COMPONENT, DIAG_WARN, 
	 "Trouble: T398 has expired on party [%d, %ld, %d]", 
	 party->sock->sclient->id, party->sock->num, party->ep_ref);
    if (party->sock->nbactp > 0) {
	send_drop_party_ack(party->sock->call_ref,
			    party->ep_ref | 0x8000,
			    ATM_CV_TIMER_EXP); /* @@@ ? */
	party->conn_timer = NULL;
	SEND_ERROR(party->sock->sclient, party->sock->num, 
		   party->ep_ref, -ETIMEDOUT);
	free_party(party);
	return;
    } else {
	send_release(party->sock->call_ref, ATM_CV_TIMER_EXP, 398); /* @@@ ? */
	party->sock->conn_timer = NULL;
	START_TIMER(party->sock, T308_1);
	new_state(party->sock, ss_rel_req);
    }
}


void on_T399(void *user) /* ps_ADD_INIT */
{
    PARTY *party = user;

    diag(COMPONENT, DIAG_DEBUG, "T399 on party [%d, %ld, %d]", 
	 party->sock->sclient->id, party->sock->num, party->ep_ref);
    if (party->pstate != ps_add_init) 
	diag(COMPONENT, DIAG_WARN, 
	     "Timer T399 expired in incompatible state %d", party->pstate);
    SEND_ERROR(party->sock->sclient, party->sock->num, party->ep_ref, 
	       -ETIMEDOUT);
    party->conn_timer = NULL;
    free_party(party);
}
