/* trace.c - Support functions for message tracing */
 
/* Written 1996-1998 by Werner Almesberger, EPFL-LRC/ICA */
 
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>

#include "../linux-atm.h"

#include "atm.h"
#include "atmd.h"
#include "atmsap.h"
#include "trace.h"
#include "uni.h"

#define DUMP_MODE
#include <q.out.h>
#include "qlib.h"


typedef struct _entry {
    int number;
    int scl_id;             /* socket descriptor to communicate with sclient */
    struct timeval time;
    void (*print)(void *msg, int size);
    const char *comment;
    void *msg;
    int size;
    struct _entry *next;
} ENTRY;


int trace_size = 0;

static int current_size = 0;
static int sequence = 0;
static ENTRY *first = NULL, *last = NULL;
static char *string = NULL;
static int curr_len;
static int new_line;


static inline void append_chunk(const char *str, int len)
{
    if (!string)
	curr_len = 0;
    if (!(string = realloc(string, curr_len + len + 1))) {
	perror("realloc");
	exit(1);
    }
    memcpy(string + curr_len, str, len);
    curr_len += len;
    string[curr_len] = 0;            /* perhaps '\0' rather...see later */
}


static void vappend(const char *fmt, va_list ap)
{
    const char *walk, *next;

    for (walk = next = fmt; *walk; walk++)
	if (*walk == '%') {
	    if (walk != next)
		append_chunk(next, walk-next);
	    if (*++walk == 's') {
		const char *str;

		str = va_arg(ap, const char *);
		append_chunk(str, strlen(str));
	    } else {
		char buf[21]; /* big enough for 64 bits */
		int num;

		while (isdigit(*walk) || *walk == 'l')
		    walk++; /* @@@ FIXME */
		if (*walk != 'd' && *walk != 'x') {
		    fprintf(stderr, "bad format character %c (%d)\n",
			    *walk, *walk);
		    exit(1);
		}
		num = va_arg(ap, int);
		sprintf(buf, *walk == 'd' ? "%d" : "%x", num);
		append_chunk(buf, strlen(buf));
	    }
	    next = walk+1;
	}
    if (walk != next)
	append_chunk(next, walk - next);
}


static void append(const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    vappend(fmt, ap);
    va_end(ap);
}


static void print_text(void *msg, int size)
{
    append("  %s\n", msg);
}


static void append_svc(const struct svc_addr *svc)
{
    char buffer[MAX_ATM_ADDR_LEN + 1];

    if (atm2text_svc(buffer, MAX_ATM_ADDR_LEN + 1, svc,
		     A2T_NAME | A2T_LOCAL | A2T_PRETTY) < 0)
	strcpy(buffer, "<invalid>");
    append("%s\n", buffer);
}


static void append_sap(const struct atm_sap *sap)
{
    char buffer[MAX_ATM_SAP_LEN + 1];

    if (sap2text(buffer, MAX_ATM_SAP_LEN + 1, sap, S2T_NAME | S2T_LOCAL) < 0)
	strcpy(buffer, "<invalid>");
    append("%s\n", buffer);
}

static void append_qos(const struct atm_qos *qos)
{
    char buffer[MAX_ATM_QOS_LEN + 1];

    if (qos2text(buffer, MAX_ATM_QOS_LEN + 1, qos, 0) < 0)
	strcpy(buffer, "<invalid>");
    append("%s\n", buffer);
}


#ifdef nomore

static void append_tp(const struct atm_trafprm *tp)
{
    static const char *name[] =
	{ "NONE", "UBR", "CBR", "VBR", "ABR", "ANYCLASS" };

    append("%s, PCR(%d) %d..%d, CDV %d, SDU %d\n",
	   tp->traffic_class < sizeof(name) / sizeof(*name) ?
		 name[tp->traffic_class] : "???",
	   tp->pcr, tp->min_pcr, tp->max_pcr, tp->max_cdv, tp->max_sdu);
}
#endif

static void qd_vdump(const char *msg, va_list ap)
{
    if (new_line)
	append("  ");
    vappend(msg, ap);
    if (string && curr_len)
	new_line = string[curr_len - 1] == '\n';
    else
	new_line = 1;
}


void qd_dump(const char *msg, ...)
{
    va_list ap;

    va_start(ap, msg);
    qd_vdump(msg, ap);
    va_end(ap);
}


void qd_report(int severity, const char *msg, ...)
{
    va_list ap;

    if (severity > Q_ERROR)
	return;
    va_start(ap, msg);
    qd_vdump(msg, ap);
    va_end(ap);
    qd_dump("\n");
}


static void print_uni(void *msg, int size)
{
    Q_DSC dsc;

    (void) qd_open(&dsc, msg, size);
    qd_close(&dsc);
}


static void print_sclient(void *msg, int size)
{
    static const char *type[] = { "as_catch_null","as_bind","as_connect",
      "as_add","as_accept","as_reject","as_listen","as_okay","as_error",
      "as_indicate","as_drop","as_close","as_itf_notify","as_modify",
      "as_identify","as_terminate" };
    struct atmsvc_msg *m = msg;

    append("  %s (vcc 0x%x, listen_vcc 0x%x, ep_ref 0x%x)\n",
	   m->type < sizeof(type) / sizeof(*type) ? type[m->type] : "???", 
	   m->vcc, m->listen_vcc, m->ep_ref);
    append("  reply %d", m->reply);
    if (m->reply) {
	const char *error;

	error = strerror(m->reply > 0 ? m->reply : -m->reply);
	append(" (%s)", error ? error : "???");
    }
    append(", aal %d\n", m->qos.aal);
    append("  pvc %s.%d.%d\n",
	   m->pvc.snatm_if, m->pvc.snatm_vpi, m->pvc.snatm_vci);
    append("  local ");
    append_svc(&m->local);
    append("  qos ");
    append_qos(&m->qos);
    append("  svc ");
    append_svc(&m->svc);
    append("  sap ");
    append_sap(&m->sap);
}


static void store(int scl_id, void (*print)(void *msg, int size), 
		  const char *comment, void *msg, int size)
{
    ENTRY *entry;

    entry = alloc_t(ENTRY);
    (void) gettimeofday(&entry->time, NULL);
    entry->number = sequence++;
    entry->scl_id = scl_id; /* socket descriptor to communicate with sclient */
    entry->print = print;
    entry->comment = comment;
    entry->msg = msg;
    entry->size = size;
    entry->next = NULL;
    if (current_size < trace_size)
	current_size++;
    else { /* handle trace_size < 1 (< 0) too */
	ENTRY *next;

	next = first->next;
	free(first->msg);
	free(first);
	if (first == last)
	    last = NULL; /* someone set trace_size < 2 */	
	first = next;
    }
    if (last)
	last->next = entry;
    else
	first = entry;
    last = entry;
}


void trace_msg(const char *msg)
{
    char *buf;

    if (!trace_size)
	return;
    buf = alloc(strlen(msg) + 1);
    strcpy(buf, msg);
    store(0, &print_text, "MESSAGE", buf, strlen(msg));
}


void trace_uni(const char *comment, const void *msg, int size)
{
    void *buf;

    if (!trace_size)
	return;
    buf = alloc(size);
    memcpy(buf, msg, size);
    store(0, &print_uni, comment, buf, size);
}


void trace_sclient(const char *comment, int scl_id, 
		   const struct atmsvc_msg *msg)
{
    struct atmsvc_msg *buf;

    if (!trace_size)
	return;
    buf = alloc_t(struct atmsvc_msg);
    *buf = *msg;
    store(scl_id, &print_sclient, comment, buf, sizeof(*msg));
}


char *get_trace(void)
{
    ENTRY *walk;

    if (string) {
	free(string);
	string = NULL;
    }
    for (walk = first; walk; walk = walk->next) {
	append("scl_id %d, number %6d, time (%d.%06d) %s:\n",
	       walk->scl_id, walk->number,
	       walk->time.tv_sec, walk->time.tv_usec, 
	       walk->comment);
	new_line = 1;
	walk->print(walk->msg, walk->size);
    }
    return string;
}
