/* uni.c - Processing of incoming UNI signaling messages */
 
/* Written 1995-1998 by Werner Almesberger, EPFL-LRC/ICA */
/* Modified 1997-1998 by Mohsen.Souissi@inria.fr, INRIA */
 
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"
#include "uni.h"
#include <q.out.h>
#include "qlib.h"

#include "proto.h"
#include "sap.h"
#include "io.h"
#include "policy.h"
#include "timeout.h"
#include "trace.h"


#define COMPONENT "UNI"


extern const char *cause_text[]; /* from mess.c */


static Q_DSC in_dsc;
static TIMER *t309 = NULL;


int send_call_proceeding(SOCKET *sock, PARTY *party)
{
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, ATM_MSG_CALL_PROC);
    q_assign(&dsc, QF_call_ref, sock->call_ref);
    if (net && (party == NULL)) { /* only in p2p conf */
	if (allocate_ci) {
	    int vci;

	    vci = get_vci(0);
	    if (vci < 0) {
		(void) q_close(&dsc);
		return vci;
	    }
	    strncpy(sock->pvc.snatm_if, signaling_pvc.snatm_if, IFNAMSIZ);
	    sock->pvc.snatm_vpi = 0;
	    sock->pvc.snatm_vci = vci;
	}
	q_assign(&dsc, QF_vpi, sock->pvc.snatm_vpi);
	q_assign(&dsc, QF_vci, sock->pvc.snatm_vci);
    }
    if (party != NULL)
	q_assign(&dsc, QF_ep_ref, party->ep_ref | 0x8000);
    if ((size = q_close(&dsc)) >= 0) 
	to_signaling(q_buffer, size);
    return 0;
}


static void setup_call(u_long call_ref, short ep_ref)
{
    SOCKET *sock = NULL, *this, *temp, **walk;
    SCLIENT *scl = NULL;
    struct svc_addr in_addr;
    struct atm_sap in_sap;
    struct atm_qos in_qos;
    int i;
    u_long num;
    PARTY *party = NULL;

    if (ep_ref <= 0)		/* p2p or first leaf conf */
	sap_decode(&in_dsc, &in_addr, &in_sap, &in_qos);
    else
	sap_decode(&in_dsc, &in_addr, &in_sap, NULL);
    if (!atmsvc_addr_in_use(in_addr)) {
	send_release_complete(call_ref, ATM_CV_UNALLOC);
	return;
    }
    if (!allow(&in_addr, ACL_IN)) {
	send_release_complete(call_ref, ATM_CV_REJ_CLIR);
	return;
    }
    temp = alloc_t(SOCKET);
    memset(temp, 0, sizeof(SOCKET));
    if (ep_ref <= 0)
	sock = lookup_sap(&in_addr, &in_sap, &in_qos,
			  &temp->local, &temp->sap, &temp->qos, 0);
    else
	sock = lookup_sap(&in_addr, &in_sap, NULL,
			  &temp->local, &temp->sap, NULL, 0);
    if (!sock) {
	free(temp);
	send_release_complete(call_ref, ATM_CV_INCOMP_DEST);
	return;
    }
    scl = sock->sclient;
    if (scl->state == DYING) {    /* later: ?????? */  
	free(temp);
	if (sock->state == ss_wait_rel) 
	    STOP_TIMER(sock);
	free_sock(sock);
	return;
    }
    if ((num = scl->connections->num + 1) == 0xffffff) 
	num = 1;
    this = new_sock(scl, num, (ep_ref < 0) ? p2p : leaf);
    this->local = temp->local;
    this->sap = temp->sap;
    if (ep_ref <= 0) {
	this->qos = temp->qos;
	this->qos.aal = ATM_AAL5;	/* hack @@@ */
    }
    free(temp);
    this->state = net && allocate_ci ? ss_proceeding : ss_indicated;
    this->call_state = cs_in_proc;
    this->call_ref = call_ref;
    if (ep_ref >= 0) {
	party = new_party(this, (u_short) ep_ref, ps_add_recv);
	this->nbactp = 1;
    }
#ifdef CISCO
    else
#endif
    if (net && allocate_ci) {
	int error;

	error = send_call_proceeding(this, party);
	if (error) {
	    free_sock(this);
	    send_release_complete(call_ref, ATM_CV_NO_CI);
	    return;
	}
    }
    /* if (sock->local) this->local = sock->local; ??? */
    diag(COMPONENT, DIAG_DEBUG, "AAL type %ld", q_fetch(&in_dsc, QF_aal_type));
    if (!net) {		/* already set by send_call_proceeding */
	int vpci;

	vpci = q_fetch(&in_dsc, QF_vpi);
	strncpy(this->pvc.snatm_if, get_itf(&vpci), IFNAMSIZ);
	this->pvc.snatm_vpi = vpci;
	this->pvc.snatm_vci = q_fetch(&in_dsc, QF_vci);
    }
    diag(COMPONENT, DIAG_DEBUG,
	 "<CR = 0x%x> ITF.VPI.VCI: %s.%d.%d", call_ref, 
	 this->pvc.snatm_if, this->pvc.snatm_vpi, this->pvc.snatm_vci);
    if (ep_ref >= 0)
	diag(COMPONENT, DIAG_DEBUG, " <EPR = %d>", party->ep_ref);      
    if (q_present(&in_dsc, QF_cgpn)) {       /* should handle E.164 too */
	char buffer[MAX_ATM_ADDR_LEN + 1];
	int plan;

	plan = q_fetch(&in_dsc, QF_cgpn_plan);
	switch (plan) {
	    case ATM_NP_AEA:
		i = q_read(&in_dsc, QF_cgpn, (void *) this->remote.prv,
			   ATM_ESA_LEN);
		break;
	    case ATM_NP_E164:
		i = q_read(&in_dsc, QF_cgpn, (void *) this->remote.pub,
			   ATM_E164_LEN);
		break;
	    default:
		diag(COMPONENT, DIAG_WARN,
		     "Ignoring cgpn with unrecognized numbering plan 0x%x\n",
		     plan);
		i = 0;
	}
	if (i) {
	    if (atm2text_svc(buffer, MAX_ATM_ADDR_LEN + 1,
			     &this->remote, pretty) < 0) 
		strcpy(buffer, "<invalid address>");
	    diag(COMPONENT, DIAG_DEBUG, "Incoming call from %s", buffer);
	}
    }
    if (ep_ref >= 0) {
	party->remote = this->remote;
	party->sap = this->sap;
    }
    send_sclient(scl, 0, num, ep_ref, as_indicate, 0, &this->pvc,
		 &this->remote, &in_addr, &this->sap, &this->qos);
    for (walk = &sock->listen; *walk; walk = &(*walk)->listen)
	continue;
    *walk = this;
    diag(COMPONENT, DIAG_DEBUG, "SE vpi.vci=%d.%d",
	 this->pvc.snatm_vpi, this->pvc.snatm_vci);
}


static void send_status(SOCKET *sock, u_long call_ref, PARTY *party, 
			u_char cause, ...)
{
    va_list ap;
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, ATM_MSG_STATUS);
    if (sock) {
	q_assign(&dsc, QF_call_ref, sock->call_ref);
	q_assign(&dsc, QF_call_state, (int) sock->call_state);
	if (party != NULL) {	/* @@@ see after */
	    q_assign(&dsc, QF_ep_ref,
		     party->ep_ref | (sock->conf == root ? 0 : 0x8000));
	    q_assign(&dsc, QF_ep_state, party->pstate);
	}
    } else {			/* see later for ep_ref @@@ */
	q_assign(&dsc, QF_call_ref, call_ref);
	q_assign(&dsc, QF_call_state, 0); /* U0 - Null / REST 0 - Null */
    }
    q_assign(&dsc, QF_cause, cause);
    va_start(ap, cause);
    switch (cause) {
	case ATM_CV_UNKNOWN_MSG_TYPE:
	case ATM_CV_INCOMP_MSG:
	    q_assign(&dsc, QF_bad_msg_type, va_arg(ap, u_char));
	    break;
	case ATM_CV_MAND_IE_MISSING:
	case ATM_CV_INVALID_IE:
	    {
		unsigned char ie;

		ie = va_arg(ap, u_char);
		q_write(&dsc, QF_ie_id6, &ie, 1);
		break;
	    }
	default:
	    ;
    }
    va_end(ap);
    if ((size = q_close(&dsc)) >= 0)
	to_signaling(q_buffer, size);
}


static void send_status_enq(SOCKET *sock, PARTY *party)
{
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, ATM_MSG_STATUS_ENQ);
    q_assign(&dsc, QF_call_ref, sock->call_ref);
    if (party != NULL)
	q_assign(&dsc, QF_ep_ref,
		 party->ep_ref | (sock->conf == root ? 0 : 0x8000));
    if ((size = q_close(&dsc)) >= 0) 
	to_signaling(q_buffer, size);
    /* see later @@@ should start T322 */
}


static void send_connect_ack(SOCKET *sock)
{
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer,MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type,ATM_MSG_CONN_ACK);
    q_assign(&dsc, QF_call_ref, sock->call_ref);
    if ((size = q_close(&dsc)) >= 0)
	to_signaling(q_buffer, size);
}


static void send_restart_ack(int vpi, int vci)
{
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, ATM_MSG_REST_ACK);
    q_assign(&dsc, QF_call_ref, 0);
    if (!vpi && !vci)
	q_assign(&dsc, QF_rst_class, ATM_RST_ALL_VC);
    else {
	q_assign(&dsc, QF_rst_class, ATM_RST_IND_VC);
	q_assign(&dsc, QF_vpi, vpi);
	q_assign(&dsc, QF_vci, vci);
    }
    if ((size = q_close(&dsc)) >= 0)
	to_signaling(q_buffer, size);
}

#ifdef Q2963_1

static void send_conn_avail(SOCKET *sock)
{
    Q_DSC dsc;
    int size;

    q_create(&dsc, q_buffer, MAX_Q_MSG);
    q_assign(&dsc, QF_msg_type, ATM_MSG_CONN_AVAIL);
    q_assign(&dsc, QF_call_ref, sock->call_ref);
    if ((size = q_close(&dsc)) >= 0)
	to_signaling(q_buffer, size);
}

#endif


static void process_uni(u_char mid, SOCKET *sock, PARTY *party)
{
    char buffer[MAX_ATM_ADDR_LEN + 1];
    int error;
    u_char cause;
      
    if (sock->sclient->state == DYING) {  
	diag(COMPONENT, DIAG_WARN, 
	     "sclient %d (DYING) is receiving a message %d",
	     sock->sclient->id, mid);
	if (sock->state == ss_wait_rel) 
	    STOP_TIMER(sock);
	free_sock(sock);
	return;
    }
    switch (mid) {
      /* 5.5.6.12 */
      case ATM_MSG_STATUS:
	{
	CALL_STATE state;
	PARTY_STATE pstate;
      
	/*
	 * NOTE: T322 isn't implemented yet, but when it is, make sure
	 *	 to only stop it on STATUS iff the cause is
	 *	 ATM_CV_RESP_STAT_ENQ. Supplementary services break if
	 *	 you stop on any STATUS.
	 */
	if (party == NULL) {
	    state = q_fetch(&in_dsc, QF_call_state);
	    if (state == cs_null)
		break; /* clear call */
	    if ((sock->call_state == cs_rel_req) ||
		(sock->call_state == cs_rel_ind))
		return;
	    if (state != sock->call_state)
		diag(COMPONENT, DIAG_WARN, "STATUS %s received in state %s",
		     cs_name[state], cs_name[sock->call_state]);
	} else {
	    pstate = q_fetch(&in_dsc, QF_ep_state);
	    if (pstate == ps_null)
		break; /* clear call */
#ifdef notyet
	    if ((party->pstate == ps_drop_init) ||
		(party->pstate == ps_drop_recv))
		return;
#endif
	    if (pstate != party->pstate)
		diag(COMPONENT, DIAG_WARN, "STATUS %s received in state %s",
		     ps_name[pstate], ps_name[party->pstate]);
	}
	}
	return;
      default:
	;
    }

    switch (mid) {		/* yet another switch block */

      /* CONNECTING, WAIT_REL, REL_REQ */
      case ATM_MSG_CALL_PROC:
	if (sock->state == ss_wait_rel || sock->state == ss_rel_req) {
	    send_status(sock, 0, NULL, ATM_CV_INCOMP_MSG, ATM_MSG_CALL_PROC);
	    return;
	}
	if (sock->state != ss_connecting)
	    break;
	/* check for 2nd CALL_PROC @@@ */
	STOP_TIMER(sock);
	if (q_present(&in_dsc, QG_conn_id)) {
	    int vpci;

	    vpci = q_fetch(&in_dsc, QF_vpi);
	    strncpy(sock->pvc.snatm_if, get_itf(&vpci), IFNAMSIZ);
	    sock->pvc.snatm_vpi = vpci;
	    sock->pvc.snatm_vci = q_fetch(&in_dsc, QF_vci);
	    diag(COMPONENT, DIAG_DEBUG, "ITF.VPI.VCI: %s.%d.%d",
		 sock->pvc.snatm_if, sock->pvc.snatm_vpi,
		 sock->pvc.snatm_vci);
	}
	START_TIMER(sock, T310);
	sock->call_state = cs_out_proc;
	return;

      /* CONNECTING, REL_REQ, p2p or first party of p2m */
      case ATM_MSG_CONNECT:
	if (sock->state == ss_rel_req) {
	    send_status(sock, 0, NULL, ATM_CV_INCOMP_MSG, ATM_MSG_CONNECT);
	    return;
	}
	if (sock->state != ss_connecting)
	    break;
	STOP_TIMER(sock);
	if (q_present(&in_dsc, QG_conn_id)) {
	    int vpci;

	    vpci = q_fetch(&in_dsc, QF_vpi);
	    strncpy(sock->pvc.snatm_if, get_itf(&vpci), IFNAMSIZ);
	    sock->pvc.snatm_vpi = vpci;
	    sock->pvc.snatm_vci = q_fetch(&in_dsc, QF_vci);
	    diag(COMPONENT, DIAG_DEBUG, "ITF.VPI.VCI: %s.%d.%d",
		 sock->pvc.snatm_if, sock->pvc.snatm_vpi,
		 sock->pvc.snatm_vci);
	}
	error = 0;
	if ((!sock->pvc.snatm_vpi && !sock->pvc.snatm_vci) ||
	    ((party != NULL) && (party->ep_ref != 0)))
	    error = -ENOPROTOOPT;
	/* more problems */
	if (error) {
	    set_error(sock, NULL, error);
	    if ((party != NULL) && (party->ep_ref != 0)) {
		/* 1st party in p2m */
		diag(COMPONENT, DIAG_WARN,
		     "<EPR = %d> bad value (must be = 0)", party->ep_ref); 
	    }
	    /* @@@ cause follows reason ??? */
	    send_release(sock->call_ref, 0); 
	    START_TIMER(sock, T308_1);
	    new_state(sock, ss_rel_req);
	    return;
	}
	send_connect_ack(sock);
	/* @@@ fill in sock->remote */
	/* @@@ fill in traffic parameters */  /* see later */
	/* original is as_okay */
	send_sclient(sock->sclient, sock->num, 0, party ? party->ep_ref : -1, 
		     as_connect, 0, &sock->pvc, NULL,
		     &sock->local, &sock->sap, &sock->qos);
	new_state(sock, ss_connected);
	if (party != NULL) {
	    party->pstate = ps_active;
	    diag(COMPONENT, DIAG_INFO, 
		 "Active open succeeded (<CR = 0x%06X>, party [%d, %ld, %d])", 
		 sock->call_ref, sock->sclient->id, sock->num, party->ep_ref);
	} else {
	    if (atm2text_svc(buffer, MAX_ATM_ADDR_LEN + 1,
			     &sock->local, 0) < 0) 
		strcpy(buffer, "<invalid>");
	    diag(COMPONENT, DIAG_INFO, 
		 "Active open succeeded (<CR = 0x%06X>, "
		 "sock [%d, %ld], addr %s)",
		 sock->call_ref, sock->sclient->id, sock->num, buffer);
	}
	return;
    
      case ATM_MSG_ADD_PARTY_ACK:
	if (party != NULL)
	    switch (party->pstate) {
	      case ps_add_init:
		STOP_TIMER(party);
		send_sclient(sock->sclient, sock->num, 0,
			     party->ep_ref, as_add, 
			     0, NULL, NULL, NULL, NULL, NULL);
		party->pstate = ps_active;
		return;
	      default:
		diag(COMPONENT, DIAG_WARN,
		     "UNI message %s is incompatible with ep_state %s", 
		     mid2name(mid), ps_name[party->pstate]);
		return;
	    }
	else
	    diag(COMPONENT, DIAG_WARN, 
		 "No matching party for UNI message %s", mid2name(mid));
	return;
    
      case ATM_MSG_DROP_PARTY:
	cause = q_fetch(&in_dsc, QF_cause);
	diag(COMPONENT, DIAG_DEBUG, "Cause %d (%s)", cause, 
	     cause > 127 ? "invalid cause" : cause_text[cause]);
    
	if (party != NULL) {
	    if (sock->state != ss_connected)
		break;
	    switch (party->pstate) {
	      case ps_active:
		party->pstate = ps_drop_recv;
		if (--sock->nbactp == 0) {
		    free_party(party);
		    send_close_sdu(sock);
		    /* Active close (as if) */
		    diag(COMPONENT,DIAG_INFO,"Active close <CR = 0x%06X>",
			 sock->call_ref);
		    send_release(sock->call_ref,
#if defined(UNI31) || defined(UNI40)
				 ATM_CV_NORMAL_CLEAR
#else
				 ATM_CV_NORMAL_UNSPEC
#endif
				 );
		    START_TIMER(sock,T308_1);
		    new_state(sock, ss_rel_req);
		    return;
		} else {
		    send_drop_party_ack(sock->call_ref, party->ep_ref, cause);
		    send_drop_sdu(party);
		    free_party(party);
		}
		return;
	      default:
		diag(COMPONENT, DIAG_WARN,
		     "UNI message %s is incompatible with ep_state %s", 
		     mid2name(mid), ps_name[party->pstate]);
		return;
	    }
	} else
	    diag(COMPONENT, DIAG_WARN,
		 "No matching party for UNI message %s", mid2name(mid));
	return;

      case ATM_MSG_ADD_PARTY_REJ:
	cause = q_fetch(&in_dsc, QF_cause);
	diag(COMPONENT, DIAG_DEBUG, "Cause %d (%s)", cause, 
	     cause > 127 ? "invalid cause" : cause_text[cause]);
	if (party != NULL)
	    switch (party->pstate) {
	      case ps_add_init:
		set_error(sock, party, -ECONNREFUSED);
		send_drop_party_ack(sock->call_ref, party->ep_ref, cause); 
		party->pstate = ps_drop_recv;
		STOP_TIMER(party);
		send_drop_sdu(party);
		return;
	      default:
		diag(COMPONENT, DIAG_WARN,
		     "UNI message %s is incompatible with ep_state %s", 
		     mid2name(mid), ps_name[party->pstate]);
		return;
	    }
	else
	    diag(COMPONENT, DIAG_WARN,
		 "No matching party for UNI message %s", mid2name(mid));
	return;

      /* ACCEPTING, WAIT_REL, REL_REQ, p2p or leaf of p2m */
      case ATM_MSG_CONN_ACK:
	if (party != NULL)
	    diag(COMPONENT, DIAG_DEBUG, "CA <CR, EPR> =<0x%06X, %d>",
		 sock->call_ref, party->ep_ref);
	else
	    diag(COMPONENT, DIAG_DEBUG, "CA vpi.vci=%d.%d",
		 sock->pvc.snatm_vpi, sock->pvc.snatm_vci);
	if (sock->state == ss_wait_rel || sock->state == ss_rel_req) {
	    send_status(sock, 0, party, ATM_CV_INCOMP_MSG, ATM_MSG_CONN_ACK);
	    return;
	}
	if (sock->state != ss_accepting) 
	    break;
	if (party != NULL)
	    if (party->pstate != ps_add_recv)
		break;
	if (sock->conf == p2p)	/* no timers when conf == leaf */
	    STOP_TIMER(sock);
	send_sclient(sock->sclient, sock->num, 0, party ? party->ep_ref : -1, 
		     as_accept, 0, NULL, NULL, &sock->local, &sock->sap, NULL);
	new_state(sock, ss_connected);
	if (party != NULL) {
	    party->pstate = ps_active;
	    diag(COMPONENT, DIAG_INFO, 
		 "Passive open party succeeded <CR = 0x%06X> [%d, %ld, %d])", 
		 sock->call_ref, sock->sclient->id, sock->num, party->ep_ref);
	} else {
	    if (atm2text_svc(buffer, MAX_ATM_ADDR_LEN + 1, 
			     &sock->local, 0) < 0) 
		strcpy(buffer, "<invalid>");
	    diag(COMPONENT, DIAG_INFO, 
		 "Passive open succeeded <CR = 0x%06X>, "
		 "sock [%d, %ld], addr %s", 
		 sock->call_ref, sock->sclient->id, sock->num,
		 buffer);
	}
	return;
    
      /* all states */
      case ATM_MSG_RELEASE:
	cause = q_fetch(&in_dsc, QF_cause);
	diag(COMPONENT, DIAG_DEBUG, "Cause %d (%s)", cause, 
	     cause > 127 ? "invalid cause" : cause_text[cause]);
	if (sock->parties != NULL)	/* dropping all parties if any */
	    free_all_parties(sock);
	switch (sock->state) {
	  case ss_connecting:		/* p2p or root */
	    set_error(sock, NULL, -ECONNREFUSED);
	    send_release_complete(sock->call_ref, 0);
	    SEND_ERROR(sock->sclient, sock->num, -1, sock->error);
	    STOP_TIMER(sock);
	    free_sock(sock);
	    return;
	  case ss_accepting:		/* p2p or leaf */
	    set_error(sock, NULL, -ECONNRESET); /* ERESTARTSYS ? */
	    send_release_complete(sock->call_ref, 0);
	    SEND_ERROR(sock->sclient, sock->num, -1, sock->error);
	    STOP_TIMER(sock);
	    free_sock(sock);
	    return;
	  case ss_rel_req:		/* p2p, root or leaf */
	    send_close_sdu(sock);
	    /* fall through */
	  case ss_wait_rel:		/* p2p, root or leaf */
	    STOP_TIMER(sock);
	    free_sock(sock);
	    return;
#ifdef Q2963_1
	  case ss_mod_req:
	    STOP_TIMER(sock);
	    /* fall through */
	  case ss_mod_lcl:
	  case ss_mod_rcv:
	  case ss_mod_fin_ok:
	  case ss_mod_fin_fail:
	  case ss_mod_fin_ack:
#endif
	  case ss_connected:
	    diag(COMPONENT, DIAG_INFO, 
		 "Passive close on <CR = 0x%06X>, sock [%d, %ld])",
		 sock->call_ref, sock->sclient->id, sock->num);
#ifdef Q2963_1
	    if (timer_handler(sock->conn_timer) == on_T361)
		STOP_TIMER(sock);
#endif
	    send_close_sdu(sock);
	    new_state(sock, ss_rel_ind);
	    return;
	  case ss_indicated:
	    /* fall through */
	  case ss_proceeding:
	    send_release_complete(sock->call_ref, 0);
	    new_state(sock, ss_zombie);
	    /* fall through */
	  case ss_rel_ind:
	    return;
	  default:
	    /* @@@ should be ATM_CV_INCOMP_MSG */
	    send_release_complete(sock->call_ref, 0);
	    break;
	}
	break;
    
      case ATM_MSG_DROP_PARTY_ACK:
	if (q_present(&in_dsc,QF_cause) == 0)
	    cause = 0;
	else {
	    cause = q_fetch(&in_dsc, QF_cause);
	    diag(COMPONENT, DIAG_DEBUG, "Cause %d (%s)", cause,
		 cause > 127 ? "invalid cause" : cause_text[cause]);
	}
	switch (party->pstate) {
	  case ps_drop_init:
	    send_drop_sdu(party);
	    STOP_TIMER(party);
	    free_party(party);
	    return;
	  default:
	    diag(COMPONENT, DIAG_WARN,
		 "UNI message %s is incompatible with ep_state %s", 
		 mid2name(mid), ps_name[party->pstate]);
	    return;
	} /* (party == NULL) has already been tested in first_pass_uni */

      case ATM_MSG_RESTART:
	set_error(sock, NULL, -ENETRESET);
	/* fall through */

      /* fall through when clearing */
      case ATM_MSG_STATUS:
	if (party != NULL) {
	    if (q_present(&in_dsc, QF_cause) == 0)
		cause = 0;
	    else {
		cause = q_fetch(&in_dsc, QF_cause);
		diag(COMPONENT, DIAG_DEBUG, "Cause %d (%s)", cause,
		     cause > 127 ? "invalid cause" : cause_text[cause]);
	    }
	    switch (party->pstate) {
	      case ps_add_init:
	      case ps_add_recv:
		set_error(sock, party, -ECONNRESET); /* ERESTARTSYS ? */
		/* fall through */
	      case ps_drop_recv:
		send_drop_sdu(party);
		/* fall through */
	      case ps_drop_init:
		STOP_TIMER(party);
		free_party(party);
		return;
	      case ps_active:
		diag(COMPONENT, DIAG_INFO, 
		     "Passive close_party on <CR = 0x%06X>, "
		     "party [%d, %ld, %d])",
		     sock->call_ref, sock->sclient->id,
		     sock->num, party->ep_ref);
		send_drop_sdu(party);
		return;
	      default:
		break;
	    }
	    break;
	}

      /* basically any state (except LISTENING and ZOMBIE) */
      case ATM_MSG_REL_COMP:
	if (mid != ATM_MSG_REL_COMP || !q_present(&in_dsc, QF_cause))
	    cause = 0;
	else {
	    cause = q_fetch(&in_dsc, QF_cause);
	    diag(COMPONENT, DIAG_DEBUG, "Cause %d (%s)", cause,
		 cause > 127 ? "invalid cause" : cause_text[cause]);
	}
	switch (sock->state) {
	  case ss_connecting:
	    set_error(sock, NULL, cause == ATM_CV_UNALLOC ?
		      -EADDRNOTAVAIL : cause == ATM_CV_RES_UNAVAIL ||
#if defined(UNI31) || defined(UNI40)
		      cause == ATM_CV_UCR_UNAVAIL ||
#endif
		      cause == ATM_CV_NO_ROUTE_DEST ? -EHOSTUNREACH :
		      cause == ATM_CV_NUM_CHANGED ? -ECONNABORTED :
		      cause == ATM_CV_DEST_OOO ? -EHOSTDOWN :
		      -ECONNREFUSED);
	    SEND_ERROR(sock->sclient, sock->num, -1, sock->error);
	    STOP_TIMER(sock);
	    free_sock(sock);
	    return;
	  case ss_accepting:
	    set_error(sock, NULL, -ECONNRESET); /* ERESTARTSYS ? */
	    SEND_ERROR(sock->sclient, sock->num, -1, sock->error);
	    STOP_TIMER(sock);
	    free_sock(sock);
	    return;
	  case ss_rel_req:
	    send_close_sdu(sock);
	    /* fall through */
	  case ss_wait_rel:
	    STOP_TIMER(sock);
	    free_sock(sock);
	    return;
#ifdef Q2963_1
	  case ss_mod_req:
	    STOP_TIMER(sock);
	    /* fall through */
	  case ss_mod_lcl:
	  case ss_mod_rcv:
	  case ss_mod_fin_ok:
	  case ss_mod_fin_fail:
	  case ss_mod_fin_ack:
#endif
	  case ss_connected:
	    diag(COMPONENT, DIAG_INFO, 
		 "Passive close on <CR = 0x%06X>, sock [%d, %ld])",
		 sock->call_ref, sock->sclient->id, sock->num);
#ifdef Q2963_1
	    if (timer_handler(sock->conn_timer) == on_T361)
		STOP_TIMER(sock);
#endif
	    send_close_sdu(sock);
	    /* fall through */
	  case ss_rel_ind:
	    new_state(sock, ss_wait_close);
	    return;
	  case ss_indicated:
	    /* fall through */
	  case ss_proceeding:
	    new_state(sock, ss_zombie);
	    return;
	  default:
	    break;
	}
	break; /* fail */

      case ATM_MSG_ALERTING:
	/*
	 * We basically ignore this junk message, except for the connection
	 * identifier it may carry.
	 */
	if (q_present(&in_dsc, QG_conn_id)) {
	    int vpci;

	    vpci = q_fetch(&in_dsc, QF_vpi);
	    strncpy(sock->pvc.snatm_if, get_itf(&vpci), IFNAMSIZ);
	    sock->pvc.snatm_vpi = vpci;
	    sock->pvc.snatm_vci = q_fetch(&in_dsc, QF_vci);
	    diag(COMPONENT, DIAG_DEBUG, "ITF.VPI.VCI: %d.%d.%d",
		 sock->pvc.snatm_if, sock->pvc.snatm_vpi,
		 sock->pvc.snatm_vci);
	}
	return;

      case ATM_MSG_NOTIFY:
	/* silently ignore this junk */
	return;

#ifdef Q2963_1
/*
 * Buglet ahead: should actually test "call_state"
 */
      case ATM_MSG_MODIFY_REQ:
	if (sock->state != ss_connected)
	    break;
	sock->new_qos = sock->qos;
	if (q_present(&in_dsc, QF_fw_pcr_01))
	    sock->new_qos.rxtp.max_pcr = q_fetch(&in_dsc, QF_fw_pcr_01);
	if (q_present(&in_dsc, QF_bw_pcr_01))
	    sock->new_qos.txtp.max_pcr = q_fetch(&in_dsc, QF_bw_pcr_01);
	send_sclient(sock->sclient, sock->num, 0, -1, as_modify,
		     ATM_MF_INC_RSV | ATM_MF_DEC_RSV | ATM_MF_DEC_SHP,
		     NULL, NULL, NULL, NULL, &sock->new_qos);
	new_state(sock, ss_mod_rcv);
	return;

      case ATM_MSG_MODIFY_ACK:
	if (sock->state != ss_mod_req)
	    break;
	STOP_TIMER(sock);
	sock->qos = sock->new_qos;
	if (q_present(&in_dsc, QG_bbrt))
	    send_conn_avail(sock);
        send_sclient(sock->sclient, sock->num, 0, -1, as_modify,
		     ATM_MF_SET, NULL, NULL, NULL, NULL, &sock->qos);
	new_state(sock, ss_mod_fin_ok);
	return;

      case ATM_MSG_MODIFY_REJ:
	if (sock->state != ss_mod_req)
	    break;
	STOP_TIMER(sock);
	sock->error = -EAGAIN;
        send_sclient(sock->sclient, sock->num, 0, -1, as_modify,
		     ATM_MF_SET, NULL, NULL, NULL, NULL, &sock->qos);
	new_state(sock, ss_mod_fin_fail);
	return;

      case ATM_MSG_CONN_AVAIL:
	if (sock->state != ss_connected)
	    break;
	STOP_TIMER(sock);
        send_sclient(sock->sclient, sock->num, 0, -1, as_modify,
		     ATM_MF_SET, NULL, NULL, NULL, NULL, &sock->qos);
	new_state(sock, ss_mod_fin_ack);
	return;
#endif

      default:
	diag(COMPONENT, DIAG_WARN, "Bad signaling message %d", mid);
	send_status(sock, 0, NULL, ATM_CV_UNKNOWN_MSG_TYPE, mid);
	return;
    } /* see this warning later @@@ */
    diag(COMPONENT, DIAG_WARN,
	 "Signaling message %s is incompatible with state "
	 "%s/%s (%d?%d) %s %s",
	 mid2name(mid), state_name[sock->state], cs_name[sock->call_state],
	 (int) sock->state, (int) sock->call_state,
	 party ? "(ep_state = )" : "", party ? ps_name[party->pstate] : "");
    send_status(sock, 0, party, ATM_CV_INCOMP_MSG, mid);
}


void clear_all_calls(void)
{
    SOCKET *curr, *next;
    
    for (curr = sockets; curr; curr = next) {
	next = curr->next;
	if (curr->call_state != cs_null) 
	    process_uni(ATM_MSG_RESTART, curr, NULL);
    }
}


void clear_all_calls_on_T309(void)
{
    clear_all_calls();
    t309 = NULL;
}


void saal_failure(void)
{
    SOCKET *curr,*next;

    trace_msg("SAAL went down");
    for (curr = sockets; curr; curr = next) {
	next = curr->next;
	if (curr->call_state != cs_null)
	    if (curr->call_state != cs_active)
		process_uni(ATM_MSG_RESTART, curr, NULL);
	    else 
		if (!t309)
		    t309 = start_timer(T309_TIME, on_T309, NULL);
    }
}


void saal_okay(void)
{
    SOCKET *curr;
    PARTY *party;
  
    trace_msg("SAAL came up");
#ifdef THOMFLEX
    /*
     * Some versions of the Thomson Thomflex 5000 won't do any signaling before
     * they get a RESTART. Whenever SAAL comes up, this may indicate that the
     * switch got booted, so we send that RESTART. We also have to clear all
     * pending connections, which isn't that nice ... Note that the rest of the
     * RESTART state machine is not implemented, so the RESTART ACKNOWLEDGE
     * will yield a warning.
     */
    {
	Q_DSC dsc;
	int size;

	clear_all_calls();
	q_create(&dsc, q_buffer, MAX_Q_MSG);
	q_assign(&dsc, QF_msg_type, ATM_MSG_RESTART);
	q_assign(&dsc, QF_call_ref, 0);
	q_assign(&dsc, QF_rst_class, ATM_RST_ALL_VC);
	if ((size = q_close(&dsc)) >= 0)
	    to_signaling(q_buffer, size);
    }
#endif
    if (!t309)
	return;
    stop_timer(t309);
    t309 = NULL;
    for (curr = sockets; curr; curr = curr->next)
	if (curr->call_state != cs_null) { 
	    send_status_enq(curr, NULL);
	    if (curr->parties != NULL)        /* if p2m call */
		for (party = curr->parties; party != NULL; party = party->next)
		    if (party->pstate != ps_null)
			send_status_enq(curr, party);
	}
}


static void first_pass_uni(void *msg)
{
    SOCKET *curr;
    PARTY *party = NULL;
    unsigned long call_ref;
    unsigned char mid;
    short ep_ref;
  
    call_ref = q_fetch(&in_dsc, QF_call_ref) ^ 0x800000;
    mid = q_fetch(&in_dsc, QF_msg_type);
    if (mid == ATM_MSG_REST_ACK)
	return;
    if (mid == ATM_MSG_RESTART) { /* 5.5.5.2 */
	int rst_class;
    
	rst_class = q_fetch(&in_dsc, QF_rst_class);
	switch (rst_class) {
	  case ATM_RST_IND_VC:
	    {
	    int vpi, vci;
	
	    if (!q_present(&in_dsc, QG_conn_id)) {
		send_status(NULL, 0L, NULL, ATM_CV_MAND_IE_MISSING,
			    ATM_IE_CONN_ID);
		return;
	    }		/* @@@ see later */
	    vpi = q_fetch(&in_dsc, QF_vpi);
	    vci = q_fetch(&in_dsc, QF_vci);
	    for (curr = sockets; curr; curr = curr->next)
		if (curr->pvc.snatm_vpi == vpi &&
		    curr->pvc.snatm_vci == vci)
		    break;
	    if (!curr) {
		send_status(NULL, 0L, NULL, ATM_CV_INVALID_IE, ATM_IE_CONN_ID);
		return;
	    }
	    process_uni(mid, curr, NULL);
	    send_restart_ack(vpi, vci);
	    }
	    break;
	  case ATM_RST_ALL_VC:
	    clear_all_calls();
	    send_restart_ack(0, 0);
	    break;
	  default:
	    send_status(NULL, 0L, NULL, ATM_CV_INVALID_IE, ATM_IE_RESTART);
	}
	return;
    }
    if (!(call_ref & 0x7fffff)) {
	return; /* bad things happen ... @@@ */
    }
  
    for (curr = sockets; curr; curr = curr->next)
	if (curr->call_ref == call_ref) 
	    break;
    if ((curr == NULL) || (curr->call_state == cs_null)) {
	/* Call reference procedural errors 5.6.5.3.1 */
	switch(mid) {
	  case ATM_MSG_SETUP:
	  case ATM_MSG_STATUS_ENQ:
	    /* do them later */
	    break;

	  case ATM_MSG_REL_COMP:
	    return;

	  case ATM_MSG_STATUS:
	    if (q_fetch(&in_dsc, QF_call_state) != (int) cs_null)
		send_release_complete(call_ref, ATM_CV_INCOMP_MSG);
	    return;

	  case ATM_MSG_RELEASE:
	    /* call has already been closed */
	    diag(COMPONENT, DIAG_WARN,
		 "No matching call for UNI message %s <CR = 0x%06X>, "
		 " (call has already been closed)",
		 mid2name(mid), call_ref);
	    send_release_complete(call_ref, ATM_CV_INV_CR);
	    return;

	  default:
	    diag(COMPONENT, DIAG_WARN, 
		 "No matching call for UNI message %s <CR = 0x%06X>", 
		 mid2name(mid), call_ref);
	    send_release_complete(call_ref, ATM_CV_INV_CR);
	    return;
	}
    }
    /* check if it is a p2m message */
    if (q_present(&in_dsc, QF_ep_ref) != 0)
	ep_ref = q_fetch(&in_dsc, QF_ep_ref) & 0x7fff;
    else
	ep_ref = -1;

    if (ep_ref >= 0) {
	if (curr != NULL) {
	    for (party = curr->parties; party != NULL; party = party->next)
		if (party->ep_ref == (u_short) ep_ref)
		    break;
	    if (party == NULL) {	/* EPR procedural errors 5.6.5.3.2.2 */
		if ((mid != ATM_MSG_SETUP) &&
		    (mid != ATM_MSG_ADD_PARTY) &&
		    (mid != ATM_MSG_DROP_PARTY_ACK)) {
		    send_drop_party_ack(call_ref, (u_short) ep_ref, 
					ATM_CV_INV_EPR); /* cause 89 */
		    return;
		}
		if (mid == ATM_MSG_DROP_PARTY_ACK)
		    return;		/* just ignore it */
		diag(COMPONENT, DIAG_WARN, 
		     "No matching party for <CR, EPR> = <0x%06X, %d>", 
		     call_ref, ep_ref);
		return;
	    } else
		diag(COMPONENT, DIAG_DEBUG, 
		     "FROM SAAL: %s (0x%02X) <CR = 0x%06X> "
		     "for party [%d, %ld, 0x%x]",
		     mid2name(((u_char *) msg)[5]), ((u_char *)msg)[5], 
		     call_ref, curr->sclient->id, curr->num, ep_ref);
	} else
	    diag(COMPONENT, DIAG_DEBUG, 
		 "FROM SAAL: %s (0x%02X) <CR = 0x%06X> "
		 "for a new party <EPR = 0x%x>",
		 mid2name(((u_char *) msg)[5]), ((u_char *)msg)[5],
		 call_ref, ep_ref);
    } else {
	if (curr != NULL)
	    diag(COMPONENT, DIAG_DEBUG,
		 "FROM SAAL: %s (0x%02X) <CR = 0x%06X> for sock [%d, %ld]",
		 mid2name(((u_char *) msg)[5]), ((u_char *)msg)[5],
		 call_ref, curr->sclient->id, curr->num);
	else
	    diag(COMPONENT, DIAG_DEBUG, 
		 "FROM SAAL: %s (0x%02X) <CR = 0x%06X> for a new sock",
		 mid2name(((u_char *) msg)[5]), ((u_char *)msg)[5], call_ref);
    }
  
    if (mid == ATM_MSG_SETUP) {
	if (!curr) 
	    setup_call(call_ref, ep_ref);
	return;
    }
    if (mid == ATM_MSG_STATUS_ENQ) {
	send_status(curr, call_ref, party, ATM_CV_RESP_STAT_ENQ);
	return;
    }
    process_uni(mid, curr, party);
}


static void abort_call(u_char *msg, int size)
{
    SOCKET *curr;
    u_long call_ref;

    if (size < 6) {
	diag(COMPONENT, DIAG_ERROR, "message too short (%d bytes)", size);
	return;
    }
    /* hope that at least the call ref is okay ... */
    call_ref = ((msg[3] << 16) | (msg[4] << 8) | msg[5]) ^ 0x800000;
    diag(COMPONENT, DIAG_ERROR, "can't parse message - aborting the call"
	 "(CR 0x%06lx)", call_ref);
    for (curr = sockets; curr; curr = curr->next)
        if (curr->call_ref == call_ref) {
	    process_uni(ATM_MSG_RESTART, curr, NULL);
	    break;
	}
    send_release_complete(call_ref, ATM_CV_PROTOCOL_ERROR);
}


void to_uni(void *msg, int size)
{
    if (q_open(&in_dsc, msg, size) < 0) {
	abort_call(msg, size);
	return;
    }
    first_pass_uni(msg);
    if (q_close(&in_dsc) < 0)
	diag(COMPONENT, DIAG_ERROR, "q_close returned <0 in to_uni");
}
